from __future__ import annotations

import calendar
from datetime import datetime
from typing import List

from policyengine_core import periods
from policyengine_core.periods import config, helpers
from policyengine_core.periods.instant_ import Instant


class Period(tuple):
    """
    Toolbox to handle date intervals.

    A period is a triple (unit, start, size), where unit is either "month" or "year", where start format is a
    (year, month, day) triple, and where size is an integer > 1.

    Since a period is a triple it can be used as a dictionary key.
    """

    def __repr__(self) -> str:
        """
        Transform period to to its Python representation as a string.

        >>> repr(period('year', 2014))
        "Period(('year', Instant((2014, 1, 1)), 1))"
        >>> repr(period('month', '2014-2'))
        "Period(('month', Instant((2014, 2, 1)), 1))"
        >>> repr(period('day', '2014-2-3'))
        "Period(('day', Instant((2014, 2, 3)), 1))"
        """
        return "{}({})".format(
            self.__class__.__name__, super(Period, self).__repr__()
        )

    def __str__(self) -> str:
        """
        Transform period to a string.

        >>> str(period(YEAR, 2014))
        '2014'

        >>> str(period(YEAR, '2014-2'))
        'year:2014-02'
        >>> str(period(MONTH, '2014-2'))
        '2014-02'

        >>> str(period(YEAR, 2012, size = 2))
        'year:2012:2'
        >>> str(period(MONTH, 2012, size = 2))
        'month:2012-01:2'
        >>> str(period(MONTH, 2012, size = 12))
        '2012'

        >>> str(period(YEAR, '2012-3', size = 2))
        'year:2012-03:2'
        >>> str(period(MONTH, '2012-3', size = 2))
        'month:2012-03:2'
        >>> str(period(MONTH, '2012-3', size = 12))
        'year:2012-03'
        """

        unit, start_instant, size = self
        if unit == config.ETERNITY:
            return "ETERNITY"
        year, month, day = start_instant

        # 1 year long period
        if (
            unit == config.MONTH
            and size == 12
            or unit == config.YEAR
            and size == 1
        ):
            if month == 1:
                # civil year starting from january
                return str(year)
            else:
                # rolling year
                return "{}:{}-{:02d}".format(config.YEAR, year, month)
        # simple month
        if unit == config.MONTH and size == 1:
            return "{}-{:02d}".format(year, month)
        # several civil years
        if unit == config.YEAR and month == 1:
            return "{}:{}:{}".format(unit, year, size)

        if unit == config.DAY:
            if size == 1:
                return "{}-{:02d}-{:02d}".format(year, month, day)
            else:
                return "{}:{}-{:02d}-{:02d}:{}".format(
                    unit, year, month, day, size
                )

        # complex period
        return "{}:{}-{:02d}:{}".format(unit, year, month, size)

    @property
    def date(self) -> datetime.date:
        assert (
            self.size == 1
        ), '"date" is undefined for a period of size > 1: {}'.format(self)
        return self.start.date

    @property
    def days(self) -> int:
        """
        Count the number of days in period.

        >>> period('day', 2014).days
        365
        >>> period('month', 2014).days
        365
        >>> period('year', 2014).days
        365

        >>> period('day', '2014-2').days
        28
        >>> period('month', '2014-2').days
        28
        >>> period('year', '2014-2').days
        365

        >>> period('day', '2014-2-3').days
        1
        >>> period('month', '2014-2-3').days
        28
        >>> period('year', '2014-2-3').days
        365
        """
        return (self.stop.date - self.start.date).days + 1

    def intersection(self, start: Instant, stop: Instant):
        if start is None and stop is None:
            return self
        period_start = self[1]
        period_stop = self.stop
        if start is None:
            start = period_start
        if stop is None:
            stop = period_stop
        if stop < period_start or period_stop < start:
            return None
        intersection_start = max(period_start, start)
        intersection_stop = min(period_stop, stop)
        if (
            intersection_start == period_start
            and intersection_stop == period_stop
        ):
            return self
        if (
            intersection_start.day == 1
            and intersection_start.month == 1
            and intersection_stop.day == 31
            and intersection_stop.month == 12
        ):
            return self.__class__(
                (
                    "year",
                    intersection_start,
                    intersection_stop.year - intersection_start.year + 1,
                )
            )
        if (
            intersection_start.day == 1
            and intersection_stop.day
            == calendar.monthrange(
                intersection_stop.year, intersection_stop.month
            )[1]
        ):
            return self.__class__(
                (
                    "month",
                    intersection_start,
                    (
                        (intersection_stop.year - intersection_start.year) * 12
                        + intersection_stop.month
                        - intersection_start.month
                        + 1
                    ),
                )
            )
        return self.__class__(
            (
                "day",
                intersection_start,
                (intersection_stop.date - intersection_start.date).days + 1,
            )
        )

    def get_subperiods(self, unit: str) -> List["Period"]:
        """
        Return the list of all the periods of unit ``unit`` contained in self.

        Examples:

        >>> period('2017').get_subperiods(MONTH)
        >>> [period('2017-01'), period('2017-02'), ... period('2017-12')]

        >>> period('year:2014:2').get_subperiods(YEAR)
        >>> [period('2014'), period('2015')]
        """
        if helpers.unit_weight(self.unit) < helpers.unit_weight(unit):
            raise ValueError(
                "Cannot subdivide {0} into {1}".format(self.unit, unit)
            )

        if unit == config.YEAR:
            return [
                self.this_year.offset(i, config.YEAR) for i in range(self.size)
            ]

        if unit == config.MONTH:
            return [
                self.first_month.offset(i, config.MONTH)
                for i in range(self.size_in_months)
            ]

        if unit == config.DAY:
            return [
                self.first_day.offset(i, config.DAY)
                for i in range(self.size_in_days)
            ]

    def offset(self, offset: int, unit: str = None) -> "Period":
        """
        Increment (or decrement) the given period with offset units.

        >>> period('day', 2014).offset(1)
        Period(('day', Instant((2014, 1, 2)), 365))
        >>> period('day', 2014).offset(1, 'day')
        Period(('day', Instant((2014, 1, 2)), 365))
        >>> period('day', 2014).offset(1, 'month')
        Period(('day', Instant((2014, 2, 1)), 365))
        >>> period('day', 2014).offset(1, 'year')
        Period(('day', Instant((2015, 1, 1)), 365))

        >>> period('month', 2014).offset(1)
        Period(('month', Instant((2014, 2, 1)), 12))
        >>> period('month', 2014).offset(1, 'day')
        Period(('month', Instant((2014, 1, 2)), 12))
        >>> period('month', 2014).offset(1, 'month')
        Period(('month', Instant((2014, 2, 1)), 12))
        >>> period('month', 2014).offset(1, 'year')
        Period(('month', Instant((2015, 1, 1)), 12))

        >>> period('year', 2014).offset(1)
        Period(('year', Instant((2015, 1, 1)), 1))
        >>> period('year', 2014).offset(1, 'day')
        Period(('year', Instant((2014, 1, 2)), 1))
        >>> period('year', 2014).offset(1, 'month')
        Period(('year', Instant((2014, 2, 1)), 1))
        >>> period('year', 2014).offset(1, 'year')
        Period(('year', Instant((2015, 1, 1)), 1))

        >>> period('day', '2011-2-28').offset(1)
        Period(('day', Instant((2011, 3, 1)), 1))
        >>> period('month', '2011-2-28').offset(1)
        Period(('month', Instant((2011, 3, 28)), 1))
        >>> period('year', '2011-2-28').offset(1)
        Period(('year', Instant((2012, 2, 28)), 1))

        >>> period('day', '2011-3-1').offset(-1)
        Period(('day', Instant((2011, 2, 28)), 1))
        >>> period('month', '2011-3-1').offset(-1)
        Period(('month', Instant((2011, 2, 1)), 1))
        >>> period('year', '2011-3-1').offset(-1)
        Period(('year', Instant((2010, 3, 1)), 1))

        >>> period('day', '2014-1-30').offset(3)
        Period(('day', Instant((2014, 2, 2)), 1))
        >>> period('month', '2014-1-30').offset(3)
        Period(('month', Instant((2014, 4, 30)), 1))
        >>> period('year', '2014-1-30').offset(3)
        Period(('year', Instant((2017, 1, 30)), 1))

        >>> period('day', 2014).offset(-3)
        Period(('day', Instant((2013, 12, 29)), 365))
        >>> period('month', 2014).offset(-3)
        Period(('month', Instant((2013, 10, 1)), 12))
        >>> period('year', 2014).offset(-3)
        Period(('year', Instant((2011, 1, 1)), 1))

        >>> period('day', '2014-2-3').offset('first-of', 'month')
        Period(('day', Instant((2014, 2, 1)), 1))
        >>> period('day', '2014-2-3').offset('first-of', 'year')
        Period(('day', Instant((2014, 1, 1)), 1))

        >>> period('day', '2014-2-3', 4).offset('first-of', 'month')
        Period(('day', Instant((2014, 2, 1)), 4))
        >>> period('day', '2014-2-3', 4).offset('first-of', 'year')
        Period(('day', Instant((2014, 1, 1)), 4))

        >>> period('month', '2014-2-3').offset('first-of')
        Period(('month', Instant((2014, 2, 1)), 1))
        >>> period('month', '2014-2-3').offset('first-of', 'month')
        Period(('month', Instant((2014, 2, 1)), 1))
        >>> period('month', '2014-2-3').offset('first-of', 'year')
        Period(('month', Instant((2014, 1, 1)), 1))

        >>> period('month', '2014-2-3', 4).offset('first-of')
        Period(('month', Instant((2014, 2, 1)), 4))
        >>> period('month', '2014-2-3', 4).offset('first-of', 'month')
        Period(('month', Instant((2014, 2, 1)), 4))
        >>> period('month', '2014-2-3', 4).offset('first-of', 'year')
        Period(('month', Instant((2014, 1, 1)), 4))

        >>> period('year', 2014).offset('first-of')
        Period(('year', Instant((2014, 1, 1)), 1))
        >>> period('year', 2014).offset('first-of', 'month')
        Period(('year', Instant((2014, 1, 1)), 1))
        >>> period('year', 2014).offset('first-of', 'year')
        Period(('year', Instant((2014, 1, 1)), 1))

        >>> period('year', '2014-2-3').offset('first-of')
        Period(('year', Instant((2014, 1, 1)), 1))
        >>> period('year', '2014-2-3').offset('first-of', 'month')
        Period(('year', Instant((2014, 2, 1)), 1))
        >>> period('year', '2014-2-3').offset('first-of', 'year')
        Period(('year', Instant((2014, 1, 1)), 1))

        >>> period('day', '2014-2-3').offset('last-of', 'month')
        Period(('day', Instant((2014, 2, 28)), 1))
        >>> period('day', '2014-2-3').offset('last-of', 'year')
        Period(('day', Instant((2014, 12, 31)), 1))

        >>> period('day', '2014-2-3', 4).offset('last-of', 'month')
        Period(('day', Instant((2014, 2, 28)), 4))
        >>> period('day', '2014-2-3', 4).offset('last-of', 'year')
        Period(('day', Instant((2014, 12, 31)), 4))

        >>> period('month', '2014-2-3').offset('last-of')
        Period(('month', Instant((2014, 2, 28)), 1))
        >>> period('month', '2014-2-3').offset('last-of', 'month')
        Period(('month', Instant((2014, 2, 28)), 1))
        >>> period('month', '2014-2-3').offset('last-of', 'year')
        Period(('month', Instant((2014, 12, 31)), 1))

        >>> period('month', '2014-2-3', 4).offset('last-of')
        Period(('month', Instant((2014, 2, 28)), 4))
        >>> period('month', '2014-2-3', 4).offset('last-of', 'month')
        Period(('month', Instant((2014, 2, 28)), 4))
        >>> period('month', '2014-2-3', 4).offset('last-of', 'year')
        Period(('month', Instant((2014, 12, 31)), 4))

        >>> period('year', 2014).offset('last-of')
        Period(('year', Instant((2014, 12, 31)), 1))
        >>> period('year', 2014).offset('last-of', 'month')
        Period(('year', Instant((2014, 1, 31)), 1))
        >>> period('year', 2014).offset('last-of', 'year')
        Period(('year', Instant((2014, 12, 31)), 1))

        >>> period('year', '2014-2-3').offset('last-of')
        Period(('year', Instant((2014, 12, 31)), 1))
        >>> period('year', '2014-2-3').offset('last-of', 'month')
        Period(('year', Instant((2014, 2, 28)), 1))
        >>> period('year', '2014-2-3').offset('last-of', 'year')
        Period(('year', Instant((2014, 12, 31)), 1))
        """
        return self.__class__(
            (
                self[0],
                self[1].offset(offset, self[0] if unit is None else unit),
                self[2],
            )
        )

    def contains(self, other: Period) -> bool:
        """
        Returns ``True`` if the period contains ``other``. For instance, ``period(2015)`` contains ``period(2015-01)``
        """
        return self.start <= other.start and self.stop >= other.stop

    @property
    def size(self) -> int:
        """
        Return the size of the period.

        >>> period('month', '2012-2-29', 4).size
        4
        """
        return self[2]

    @property
    def size_in_months(self) -> int:
        """
        Return the size of the period in months.

        >>> period('month', '2012-2-29', 4).size_in_months
        4
        >>> period('year', '2012', 1).size_in_months
        12
        """
        if self[0] == config.MONTH:
            return self[2]
        if self[0] == config.YEAR:
            return self[2] * 12
        raise ValueError(
            "Cannot calculate number of months in {0}".format(self[0])
        )

    @property
    def size_in_days(self) -> int:
        """
        Return the size of the period in days.

        >>> period('month', '2012-2-29', 4).size_in_days
        28
        >>> period('year', '2012', 1).size_in_days
        366
        """
        unit, instant, length = self

        if unit == config.DAY:
            return length
        if unit in [config.MONTH, config.YEAR]:
            last_day = self.start.offset(length, unit).offset(-1, config.DAY)
            return (last_day.date - self.start.date).days + 1

        raise ValueError("Cannot calculate number of days in {0}".format(unit))

    @property
    def start(self) -> periods.Instant:
        """
        Return the first day of the period as an Instant instance.

        >>> period('month', '2012-2-29', 4).start
        Instant((2012, 2, 29))
        """
        return self[1]

    @property
    def stop(self) -> periods.Instant:
        """
        Return the last day of the period as an Instant instance.

        >>> period('year', 2014).stop
        Instant((2014, 12, 31))
        >>> period('month', 2014).stop
        Instant((2014, 12, 31))
        >>> period('day', 2014).stop
        Instant((2014, 12, 31))

        >>> period('year', '2012-2-29').stop
        Instant((2013, 2, 28))
        >>> period('month', '2012-2-29').stop
        Instant((2012, 3, 28))
        >>> period('day', '2012-2-29').stop
        Instant((2012, 2, 29))

        >>> period('year', '2012-2-29', 2).stop
        Instant((2014, 2, 28))
        >>> period('month', '2012-2-29', 2).stop
        Instant((2012, 4, 28))
        >>> period('day', '2012-2-29', 2).stop
        Instant((2012, 3, 1))
        """
        unit, start_instant, size = self
        year, month, day = start_instant
        if unit == config.ETERNITY:
            return periods.Instant((float("inf"), float("inf"), float("inf")))
        if unit == "day":
            if size > 1:
                day += size - 1
                month_last_day = calendar.monthrange(year, month)[1]
                while day > month_last_day:
                    month += 1
                    if month == 13:
                        year += 1
                        month = 1
                    day -= month_last_day
                    month_last_day = calendar.monthrange(year, month)[1]
        else:
            if unit == "month":
                month += size
                while month > 12:
                    year += 1
                    month -= 12
            else:
                assert unit == "year", "Invalid unit: {} of type {}".format(
                    unit, type(unit)
                )
                year += size
            day -= 1
            if day < 1:
                month -= 1
                if month == 0:
                    year -= 1
                    month = 12
                day += calendar.monthrange(year, month)[1]
            else:
                month_last_day = calendar.monthrange(year, month)[1]
                if day > month_last_day:
                    month += 1
                    if month == 13:
                        year += 1
                        month = 1
                    day -= month_last_day
        return periods.Instant((year, month, day))

    @property
    def unit(self) -> str:
        return self[0]

    # Reference periods

    @property
    def last_3_months(self) -> "Period":
        return self.first_month.start.period("month", 3).offset(-3)

    @property
    def last_month(self) -> "Period":
        return self.first_month.offset(-1)

    @property
    def last_year(self) -> "Period":
        return self.start.offset("first-of", "year").period("year").offset(-1)

    @property
    def n_2(self) -> "Period":
        return self.start.offset("first-of", "year").period("year").offset(-2)

    @property
    def this_year(self) -> "Period":
        return self.start.offset("first-of", "year").period("year")

    @property
    def first_month(self) -> "Period":
        return self.start.offset("first-of", "month").period("month")

    @property
    def first_day(self) -> "Period":
        return self.start.period("day")



import logging
from typing import Any, Dict, List, Type

from policyengine_core.enums import Enum
from policyengine_core.parameters.parameter import Parameter
from policyengine_core.parameters.parameter_node import ParameterNode
from policyengine_core.variables import Variable


def homogenize_parameter_structures(
    root: ParameterNode, variables: Dict[str, Variable], default_value: Any = 0
) -> ParameterNode:
    """
    Homogenize the structure of a parameter tree to match the structure of the variables.

    Args:
        root (ParameterNode): The root of the parameter tree.
        variables (Dict[str, Variable]): The variables to match the structure to.
        default_value (Any, optional): The default value to use for missing parameters. Defaults to 0.

    Returns:
        ParameterNode: The root of the homogenized parameter tree.
    """
    for node in root.get_descendants():
        if isinstance(node, ParameterNode):
            breakdown = get_breakdown_variables(node)
            node = homogenize_parameter_node(
                node, breakdown, variables, default_value
            )
    return root


def get_breakdown_variables(node: ParameterNode) -> List[str]:
    """
    Returns the list of variables that are used to break down the parameter.
    """
    breakdown = node.metadata.get("breakdown")
    if breakdown is not None:
        if isinstance(breakdown, str):
            # Single element, cast to list.
            breakdown = [breakdown]
        if not isinstance(breakdown, list):
            # Not a list, skip process and warn.
            logging.warning(
                f"Invalid breakdown metadata for parameter {node.name}: {type(breakdown)}"
            )
            return None
        return breakdown
    else:
        return None


def homogenize_parameter_node(
    node: ParameterNode,
    breakdown: List[str],
    variables: Dict[str, Variable],
    default_value: Any,
) -> ParameterNode:
    if breakdown is None:
        return node
    first_breakdown = breakdown[0]
    if isinstance(first_breakdown, list):
        possible_values = first_breakdown
    elif first_breakdown in variables:
        dtype = variables[first_breakdown].value_type
        if dtype == Enum:
            possible_values = list(
                map(
                    lambda enum: enum.name,
                    variables[first_breakdown].possible_values,
                )
            )
        elif dtype == bool:
            possible_values = [True, False]
    else:
        # Try to execute the breakdown as Python code
        possible_values = list(eval(first_breakdown))
    if not hasattr(node, "children"):
        node = ParameterNode(
            node.name,
            data={
                child: {"2000-01-01": default_value}
                for child in possible_values
            },
        )
    missing_values = set(possible_values) - set(node.children)
    further_breakdown = len(breakdown) > 1
    for value in missing_values:
        if str(value) not in node.children:
            # Integers behave strangely, this fixes it.
            node.add_child(
                str(value),
                Parameter(
                    node.name + "." + str(value), {"2000-01-01": default_value}
                ),
            )
    for child in node.children:
        if child.split(".")[-1] not in possible_values:
            try:
                int(child)
                is_int = True
            except:
                is_int = False
            if not is_int or str(child) not in node.children:
                logging.warning(
                    f"Parameter {node.name} has a child {child} that is not in the possible values of {first_breakdown}, ignoring."
                )
        if further_breakdown:
            node.children[child] = homogenize_parameter_node(
                node.children[child], breakdown[1:], variables, default_value
            )
    return node



from .computation_log import ComputationLog
from .flat_trace import FlatTrace
from .full_tracer import FullTracer
from .performance_log import PerformanceLog
from .simple_tracer import SimpleTracer
from .trace_node import TraceNode
from .tracing_parameter_node_at_instant import TracingParameterNodeAtInstant



from policyengine_core.errors import CycleError, NaNCreationError, SpiralError

from .helpers import (
    calculate_output_add,
    calculate_output_divide,
    check_type,
    transform_to_strict_syntax,
)
from .microsimulation import Microsimulation
from .simulation import Simulation
from .simulation_builder import SimulationBuilder



from policyengine_core.parameters.parameter import Parameter
from policyengine_core.parameters.parameter_node import ParameterNode


def get_parameter(root: ParameterNode, parameter: str) -> Parameter:
    """Gets a parameter from the tree by name.

    Args:
        root (ParameterNode): The root of the parameter tree.
        parameter (str): The name of the parameter to get.

    Returns:
        Parameter: The parameter.
    """
    node = root
    for name in parameter.split("."):
        try:
            if "[" not in name:
                node = node.children[name]
            else:
                try:
                    name, index = name.split("[")
                    index = int(index[:-1])
                    node = node.children[name].brackets[index]
                except:
                    raise ValueError(
                        "Invalid bracket syntax (should be e.g. tax.brackets[3].rate"
                    )
        except:
            raise ValueError(
                f"Could not find the parameter (failed at {name})."
            )
    return node



import random

import numpy as np

from policyengine_core.simulations import Simulation


def make_simulation(tax_benefit_system, nb_persons, nb_groups, **kwargs):
    """
    Generate a simulation containing nb_persons persons spread in nb_groups groups.

    Example:

    >>> from policyengine_core.scripts.simulation_generator import make_simulation
    >>> from openfisca_france import CountryTaxBenefitSystem
    >>> tbs = CountryTaxBenefitSystem()
    >>> simulation = make_simulation(tbs, 400, 100)  # Create a simulation with 400 persons, spread among 100 families
    >>> simulation.calculate('revenu_disponible', 2017)
    """
    simulation = Simulation(tax_benefit_system=tax_benefit_system, **kwargs)
    simulation.persons.ids = np.arange(nb_persons)
    simulation.persons.count = nb_persons
    adults = [0] + sorted(random.sample(range(1, nb_persons), nb_groups - 1))

    members_entity_id = np.empty(nb_persons, dtype=int)

    # A legacy role is an index that every person within an entity has. For instance, the 'demandeur' has legacy role 0, the 'conjoint' 1, the first 'child' 2, the second 3, etc.
    members_legacy_role = np.empty(nb_persons, dtype=int)

    id_group = -1
    for id_person in range(nb_persons):
        if id_person in adults:
            id_group += 1
            legacy_role = 0
        else:
            legacy_role = 2 if legacy_role == 0 else legacy_role + 1
        members_legacy_role[id_person] = legacy_role
        members_entity_id[id_person] = id_group

    for entity in simulation.populations.values():
        if not entity.is_person:
            entity.members_entity_id = members_entity_id
            entity.count = nb_groups
            entity.members_role = np.where(
                members_legacy_role == 0,
                entity.flattened_roles[0],
                entity.flattened_roles[-1],
            )
    return simulation



import copy
import os
import typing
from typing import Any, Iterable

from policyengine_core import commons, parameters, tools
from policyengine_core.errors import ParameterParsingError
from policyengine_core.parameters import AtInstantLike, config, helpers
from policyengine_core.periods.instant_ import Instant
from policyengine_core.taxscales import (
    LinearAverageRateTaxScale,
    MarginalAmountTaxScale,
    MarginalRateTaxScale,
    SingleAmountTaxScale,
    TaxScaleLike,
)


class ParameterScale(AtInstantLike):
    """
    A parameter scale (for instance a  marginal scale).
    """

    # 'unit' and 'reference' are only listed here for backward compatibility
    _allowed_keys = config.COMMON_KEYS.union({"brackets"})

    def __init__(self, name: str, data: dict, file_path: str):
        """
        :param name: name of the scale, eg "taxes.some_scale"
        :param data: Data loaded from a YAML file. In case of a reform, the data can also be created dynamically.
        :param file_path: File the parameter was loaded from.
        """
        self.name: str = name
        self.file_path: str = file_path
        helpers._validate_parameter(
            self, data, data_type=dict, allowed_keys=self._allowed_keys
        )
        self.description: str = data.get("description")
        self.metadata: typing.Dict = {}
        self.metadata.update(data.get("metadata", {}))

        if not isinstance(data.get("brackets", []), list):
            raise ParameterParsingError(
                "Property 'brackets' of scale '{}' must be of type array.".format(
                    self.name
                ),
                self.file_path,
            )

        brackets = []
        for i, bracket_data in enumerate(data.get("brackets", [])):
            bracket_name = helpers._compose_name(name, item_name=i)
            bracket = parameters.ParameterScaleBracket(
                name=bracket_name, data=bracket_data, file_path=file_path
            )
            brackets.append(bracket)
        self.brackets: typing.List[parameters.ParameterScaleBracket] = brackets

    def __getitem__(self, key: str) -> Any:
        if isinstance(key, int) and key < len(self.brackets):
            return self.brackets[key]
        else:
            raise KeyError(key)

    def __repr__(self) -> str:
        return os.linesep.join(
            ["brackets:"]
            + [
                tools.indent("-" + tools.indent(repr(bracket))[1:])
                for bracket in self.brackets
            ]
        )

    def get_descendants(self) -> Iterable:
        return iter(())

    def clone(self) -> "ParameterScale":
        clone = commons.empty_clone(self)
        clone.__dict__ = self.__dict__.copy()

        clone.brackets = [bracket.clone() for bracket in self.brackets]
        clone.metadata = copy.deepcopy(self.metadata)

        return clone

    def _get_at_instant(self, instant: Instant) -> TaxScaleLike:
        brackets = [
            bracket.get_at_instant(instant) for bracket in self.brackets
        ]

        if self.metadata.get("type") == "single_amount":
            scale = SingleAmountTaxScale()
            for bracket in brackets:
                if (
                    "amount" in bracket._children
                    and "threshold" in bracket._children
                ):
                    amount = bracket.amount
                    threshold = bracket.threshold
                    scale.add_bracket(threshold, amount)
            return scale
        elif any("amount" in bracket._children for bracket in brackets):
            scale = MarginalAmountTaxScale()
            for bracket in brackets:
                if (
                    "amount" in bracket._children
                    and "threshold" in bracket._children
                ):
                    amount = bracket.amount
                    threshold = bracket.threshold
                    scale.add_bracket(threshold, amount)
            return scale
        elif any("average_rate" in bracket._children for bracket in brackets):
            scale = LinearAverageRateTaxScale()

            for bracket in brackets:
                if "base" in bracket._children:
                    base = bracket.base
                else:
                    base = 1.0
                if (
                    "average_rate" in bracket._children
                    and "threshold" in bracket._children
                ):
                    average_rate = bracket.average_rate
                    threshold = bracket.threshold
                    scale.add_bracket(threshold, average_rate * base)
            return scale
        else:
            scale = MarginalRateTaxScale()

            for bracket in brackets:
                if "base" in bracket._children:
                    base = bracket.base
                else:
                    base = 1.0
                if (
                    "rate" in bracket._children
                    and "threshold" in bracket._children
                ):
                    rate = bracket.rate
                    threshold = bracket.threshold
                    scale.add_bracket(threshold, rate * base)
            return scale



from policyengine_core.errors import EmptyArgumentError

from .abstract_rate_tax_scale import AbstractRateTaxScale
from .abstract_tax_scale import AbstractTaxScale
from .amount_tax_scale_like import AmountTaxScaleLike
from .helpers import combine_tax_scales
from .linear_average_rate_tax_scale import LinearAverageRateTaxScale
from .marginal_amount_tax_scale import MarginalAmountTaxScale
from .marginal_rate_tax_scale import MarginalRateTaxScale
from .rate_tax_scale_like import RateTaxScaleLike
from .single_amount_tax_scale import SingleAmountTaxScale
from .tax_scale_like import TaxScaleLike



import os
import traceback
import typing

import numpy


class EmptyArgumentError(IndexError):
    """Exception raised when a method is called with an empty argument."""

    message: str

    def __init__(
        self,
        class_name: str,
        method_name: str,
        arg_name: str,
        arg_value: typing.Union[typing.List, numpy.ndarray],
    ) -> None:
        message = [
            f"'{class_name}.{method_name}' can't be run with an empty '{arg_name}':\n",
            f">>> {arg_name}",
            f"{arg_value}\n",
            "Here are some hints to help you get this working:\n",
            f"- Check that '{class_name}' isn't empty (see '{class_name}.add_bracket')",
            f"- Check that '{arg_name}' is being properly assigned "
            f"('{arg_name}' should be a non empty '{type(arg_value).__name__}')\n",
            "For further support, please do not hesitate to:\n",
            "- Take a look at the official documentation https://openfisca.org/doc",
            "- Open an issue on https://github.com/openfisca/openfisca-core/issues/new",
            "- Mention us via https://twitter.com/openfisca",
            "- Drop us a line to contact@openfisca.org\n",
            "😃",
        ]
        stacktrace = os.linesep.join(traceback.format_stack())
        self.message = os.linesep.join([f"  {line}" for line in message])
        self.message = os.linesep.join([stacktrace, self.message])
        super().__init__(self.message)



from typing import TypeVar

from policyengine_core.types import ArrayType

T = TypeVar("T")


def empty_clone(original: T) -> T:
    """Creates an empty instance of the same class of the original object.

    Args:
        original: An object to clone.

    Returns:
        The cloned, empty, object.

    Examples:
        >>> Foo = type("Foo", (list,), {})
        >>> foo = Foo([1, 2, 3])
        >>> foo
        [1, 2, 3]

        >>> bar = empty_clone(foo)
        >>> bar
        []

        >>> isinstance(bar, Foo)
        True

    """

    Dummy: object
    new: T

    Dummy = type(
        "Dummy",
        (original.__class__,),
        {"__init__": lambda self: None},
    )

    new = Dummy()
    new.__class__ = original.__class__
    return new


def stringify_array(array: ArrayType) -> str:
    """Generates a clean string representation of a numpy array.

    Args:
        array: An array.

    Returns:
        :obj:`str`:
        "None" if the ``array`` is None, the stringified ``array`` otherwise.

    Examples:
        >>> import numpy
        >>> stringify_array(None)
        'None'

        >>> array = numpy.array([10, 20.])
        >>> stringify_array(array)
        '[10.0, 20.0]'

        >>> array = numpy.array(["10", "Twenty"])
        >>> stringify_array(array)
        '[10, Twenty]'

        >>> array = numpy.array([list, dict(), stringify_array])
        >>> stringify_array(array)
        "[<class 'list'>, {}, <function stringify_array...]"

    """

    if array is None:
        return "None"

    return f"[{', '.join(str(cell) for cell in array)}]"



import logging
import subprocess
import warnings
from typing import Dict

from policyengine_core.data.dataset import Dataset


class PrivateDataset(Dataset):
    """Private datasets are stored on Google Cloud Buckets (requires the Google Cloud CLI to be installed)."""

    bucket_name: str
    """The name of the Google Cloud Storage bucket to use for this dataset."""
    filename_by_year: Dict[int, str] = None
    """A dictionary mapping years to filenames."""

    def _get_storage_bucket(self):
        from google.cloud import storage

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            try:
                client = storage.Client()
            except:
                logging.info(
                    "Could not automatically authenticate with Google Cloud, prompting login..."
                )
                failed_login = subprocess.check_call(
                    ["gcloud auth application-default login"],
                    shell=True,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )
                if not failed_login:
                    logging.info(
                        "Successfully logged in to Google Cloud, attempting to download..."
                    )
                    client = storage.Client()
                else:
                    raise Exception(
                        "Could not authenticate with Google Cloud."
                    )
        try:
            return client.get_bucket(self.bucket_name)
        except Exception as e:
            raise Exception(
                f"Your account does not have sufficient permissions: {e}."
            )

    def download(self, year: int):
        """Downloads the dataset from Google Cloud Storage for the given year. Overwrites any existing dataset for the year specified.

        Args:
            year (int): The year of the dataset to download.

        Raises:
            FileNotFoundError: If the dataset could not be found on Google Cloud Storage.
        """
        bucket = self._get_storage_bucket()
        filenames = list(map(lambda blob: blob.name, bucket.list_blobs()))
        filename = self.filename_by_year[year]

        if filename not in filenames:
            raise FileNotFoundError(
                f"Could not find {filename} in the {self.bucket_name} bucket."
            )

        blob = bucket.blob(filename)
        with open(self.file(year), "wb") as f:
            blob.download_to_file(f)

        logging.info("Successfully downloaded and saved dataset.")

    def upload(self, year: int, filename: str = None):
        """Uploads the dataset to Google Cloud Storage for the given year.

        Args:
            year (int): The year of the dataset to upload.
        """
        bucket = self._get_storage_bucket()
        if filename is None:
            # Overwriting existing dataset - check this is intended with a prompt.

            if year in self.filename_by_year.keys():
                overwrite = input(
                    f"Are you sure you want to overwrite the existing dataset for year {year}? (y/n): "
                )
                if overwrite.lower() != "y":
                    raise Exception("Upload cancelled.")
            filename = self.filename_by_year[year]
            overwriting_existing = True
        else:
            overwriting_existing = False

        blob = bucket.blob(filename)
        blob.upload_from_filename(self.file(year))

        logging.info("Successfully uploaded dataset.")

        if overwriting_existing:
            logging.info(
                f"You have overwritten the existing {filename} file. Anyone using this dataset will transition to the new dataset automatically when they next download the file."
            )
        else:
            logging.info(
                f"You have uploaded a new file, {filename}, to the {self.bucket_name} bucket. Users will not automatically upgrade, unless you add the filename {filename} to the `filename_by_year` dictionary in this dataset."
            )



class Reference:
    """A reference to a piece of legislation or other document."""

    label: str
    """The display text to use for the reference."""

    href: str
    """The URL to link to."""

    type: str = "general"
    """The type of reference."""



from __future__ import annotations

import typing
from typing import Dict, List, Union

if typing.TYPE_CHECKING:
    from numpy.typing import ArrayLike

    from policyengine_core.periods import Period

    Stack = List[Dict[str, Union[str, Period]]]


class SimpleTracer:

    _stack: Stack

    def __init__(self) -> None:
        self._stack = []

    def record_calculation_start(
        self, variable: str, period: Period, branch_name: str = "default"
    ) -> None:
        self.stack.append(
            {"name": variable, "period": period, "branch_name": branch_name}
        )

    def record_calculation_result(self, value: ArrayLike) -> None:
        pass  # ignore calculation result

    def record_parameter_access(
        self, parameter: str, period, branch_name: str, value
    ):
        pass

    def record_calculation_end(self) -> None:
        self.stack.pop()

    @property
    def stack(self) -> Stack:
        return self._stack



import os
import typing
import warnings

import yaml

from policyengine_core.warnings import LibYAMLWarning

try:
    from yaml import CLoader as Loader
except ImportError:
    message = [
        "libyaml is not installed in your environment.",
        "This can make OpenFisca slower to start.",
        "Once you have installed libyaml, run 'pip uninstall pyyaml && pip install pyyaml --no-cache-dir'",
        "so that it is used in your Python environment." + os.linesep,
    ]
    warnings.warn(" ".join(message), LibYAMLWarning)
    from yaml import (
        Loader,
    )  # type: ignore # (see https://github.com/python/mypy/issues/1153#issuecomment-455802270)

ALLOWED_PARAM_TYPES = (float, int, bool, type(None), typing.List)
COMMON_KEYS = {"description", "metadata", "documentation"}
FILE_EXTENSIONS = {".yaml", ".yml"}


def date_constructor(_loader, node):
    return node.value


yaml.add_constructor(
    "tag:yaml.org,2002:timestamp", date_constructor, Loader=Loader
)


def dict_no_duplicate_constructor(loader, node, deep=False):
    keys = [key.value for key, value in node.value]

    if len(keys) != len(set(keys)):
        duplicate = next((key for key in keys if keys.count(key) > 1))
        raise yaml.parser.ParserError(
            "", node.start_mark, f"Found duplicate key '{duplicate}'"
        )

    return loader.construct_mapping(node, deep)


yaml.add_constructor(
    yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
    dict_no_duplicate_constructor,
    Loader=Loader,
)



from __future__ import annotations

import typing
import warnings

from policyengine_core.taxscales.rate_tax_scale_like import RateTaxScaleLike

if typing.TYPE_CHECKING:
    import numpy

    NumericalArray = typing.Union[numpy.int_, numpy.float_]


class AbstractRateTaxScale(RateTaxScaleLike):
    """
    Base class for various types of rate-based tax scales: marginal rate,
    linear average rate...
    """

    def __init__(
        self,
        name: typing.Optional[str] = None,
        option: typing.Any = None,
        unit: typing.Any = None,
    ) -> None:
        message = [
            "The 'AbstractRateTaxScale' class has been deprecated since",
            "version 34.7.0, and will be removed in the future.",
        ]

        warnings.warn(" ".join(message), DeprecationWarning)
        super().__init__(name, option, unit)

    def calc(
        self,
        tax_base: NumericalArray,
        right: bool,
    ) -> typing.NoReturn:
        raise NotImplementedError(
            "Method 'calc' is not implemented for "
            f"{self.__class__.__name__}",
        )



from policyengine_core.parameters.parameter import Parameter
from policyengine_core.parameters.parameter_node import ParameterNode


def propagate_parameter_metadata(root: ParameterNode) -> ParameterNode:
    """Passes parameter metadata to descendents where this is specified.

    Breakdown metadata is ignored.

    Args:
        root (ParameterNode): The root node.

    Returns:
        ParameterNode: The edited parameter root.
    """

    UNPROPAGAGED_METADATA = ["breakdown"]

    for parameter in root.get_descendants():
        if parameter.metadata.get("propagate_metadata_to_children"):
            for descendant in parameter.get_descendants():
                descendant.metadata.update(
                    {
                        key: value
                        for key, value in parameter.metadata.items()
                        if key not in UNPROPAGAGED_METADATA
                    }
                )

    return root



from __future__ import annotations

import csv
import importlib.resources
import itertools
import json
import os
import typing

from .. import tracers

if typing.TYPE_CHECKING:
    Trace = typing.Dict[str, dict]
    Calculation = typing.Tuple[str, dict]
    SortedTrace = typing.List[Calculation]


class PerformanceLog:
    def __init__(self, full_tracer: tracers.FullTracer) -> None:
        self._full_tracer = full_tracer

    def generate_graph(self, dir_path: str) -> None:
        with open(os.path.join(dir_path, "performance_graph.html"), "w") as f:
            template = importlib.resources.read_text(
                "policyengine_core.scripts.assets",
                "index.html",
            )

            perf_graph_html = template.replace(
                "{{data}}",
                json.dumps(self._json()),
            )

            f.write(perf_graph_html)

    def generate_performance_tables(self, dir_path: str) -> None:
        flat_trace = self._full_tracer.get_flat_trace()

        csv_rows = [
            {
                "name": key,
                "calculation_time": trace["calculation_time"],
                "formula_time": trace["formula_time"],
            }
            for key, trace in flat_trace.items()
        ]

        self._write_csv(
            os.path.join(dir_path, "performance_table.csv"),
            csv_rows,
        )

        aggregated_csv_rows = [
            {"name": key, **aggregated_time}
            for key, aggregated_time in self.aggregate_calculation_times(
                flat_trace
            ).items()
        ]

        self._write_csv(
            os.path.join(dir_path, "aggregated_performance_table.csv"),
            aggregated_csv_rows,
        )

    def aggregate_calculation_times(
        self,
        flat_trace: Trace,
    ) -> typing.Dict[str, dict]:
        def _aggregate_calculations(calculations: list) -> dict:
            calculation_count = len(calculations)

            calculation_time = sum(
                calculation[1]["calculation_time"]
                for calculation in calculations
            )

            formula_time = sum(
                calculation[1]["formula_time"] for calculation in calculations
            )

            return {
                "calculation_count": calculation_count,
                "calculation_time": tracers.TraceNode.round(calculation_time),
                "formula_time": tracers.TraceNode.round(formula_time),
                "avg_calculation_time": tracers.TraceNode.round(
                    calculation_time / calculation_count
                ),
                "avg_formula_time": tracers.TraceNode.round(
                    formula_time / calculation_count
                ),
            }

        def _groupby(calculation: Calculation) -> str:
            return calculation[0].split("<")[0]

        all_calculations: SortedTrace = sorted(flat_trace.items())

        return {
            variable_name: _aggregate_calculations(list(calculations))
            for variable_name, calculations in itertools.groupby(
                all_calculations, _groupby
            )
        }

    def _json(self) -> dict:
        children = [self._json_tree(tree) for tree in self._full_tracer.trees]
        calculations_total_time = sum(child["value"] for child in children)

        return {
            "name": "All calculations",
            "value": calculations_total_time,
            "children": children,
        }

    def _json_tree(self, tree: tracers.TraceNode) -> dict:
        calculation_total_time = tree.calculation_time()
        children = [self._json_tree(child) for child in tree.children]

        return {
            "name": f"{tree.name}<{tree.period}, ({tree.branch_name})>",
            "value": calculation_total_time,
            "children": children,
        }

    def _write_csv(self, path: str, rows: typing.List[dict]) -> None:
        fieldnames = list(rows[0].keys())

        with open(path, "w") as csv_file:
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()

            for row in rows:
                writer.writerow(row)



from __future__ import annotations

import abc
import copy
import typing

import numpy

from policyengine_core import commons

if typing.TYPE_CHECKING:
    NumericalArray = typing.Union[numpy.int_, numpy.float_]


class TaxScaleLike(abc.ABC):
    """
    Base class for various types of tax scales: amount-based tax scales,
    rate-based tax scales...
    """

    name: typing.Optional[str]
    option: typing.Any
    unit: typing.Any
    thresholds: typing.List

    @abc.abstractmethod
    def __init__(
        self,
        name: typing.Optional[str] = None,
        option: typing.Any = None,
        unit: typing.Any = None,
    ) -> None:
        self.name = name or "Untitled TaxScale"
        self.option = option
        self.unit = unit
        self.thresholds = []

    def __eq__(self, _other: object) -> typing.NoReturn:
        raise NotImplementedError(
            "Method '__eq__' is not implemented for "
            f"{self.__class__.__name__}",
        )

    def __ne__(self, _other: object) -> typing.NoReturn:
        raise NotImplementedError(
            "Method '__ne__' is not implemented for "
            f"{self.__class__.__name__}",
        )

    @abc.abstractmethod
    def __repr__(self) -> str:
        ...

    @abc.abstractmethod
    def calc(
        self,
        tax_base: NumericalArray,
        right: bool,
    ) -> numpy.float_:
        ...

    @abc.abstractmethod
    def to_dict(self) -> dict:
        ...

    def copy(self) -> typing.Any:
        new = commons.empty_clone(self)
        new.__dict__ = copy.deepcopy(self.__dict__)
        return new



from typing import Callable, Union

import numpy

from policyengine_core.parameters.parameter_node_at_instant import (
    ParameterNodeAtInstant,
)
from policyengine_core.periods import Instant, Period
from policyengine_core.populations import GroupPopulation, Population

#: A collection of :obj:`.Entity` or :obj:`.GroupEntity`.
People = Union[Population, GroupPopulation]

#: A callable to get the parameters for the given instant.
Params = Callable[[Instant], ParameterNodeAtInstant]

#: A callable defining a calculation, or a rule, on a system.
Formula = Callable[[People, Period, Params], numpy.ndarray]



# -*- coding: utf-8 -*-

import logging
import os
import sys

from policyengine_core.scripts import build_tax_benefit_system
from policyengine_core.tools.test_runner import run_tests


def main(parser):
    args = parser.parse_args()
    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.WARNING,
        stream=sys.stdout,
    )

    tax_benefit_system = build_tax_benefit_system(
        args.country_package, args.extensions, args.reforms
    )

    options = {
        "pdb": args.pdb,
        "performance_graph": args.performance_graph,
        "performance_tables": args.performance_tables,
        "verbose": args.verbose,
        "aggregate": args.aggregate,
        "max_depth": args.max_depth,
        "name_filter": args.name_filter,
        "only_variables": args.only_variables,
        "ignore_variables": args.ignore_variables,
    }

    paths = [os.path.abspath(path) for path in args.path]
    sys.exit(run_tests(tax_benefit_system, paths, options))



from __future__ import annotations

import enum
from typing import Union

import numpy

from .config import ENUM_ARRAY_DTYPE
from .enum_array import EnumArray


class Enum(enum.Enum):
    """
    Enum based on `enum34 <https://pypi.python.org/pypi/enum34/>`_, whose items
    have an index.
    """

    # Tweak enums to add an index attribute to each enum item
    def __init__(self, name: str) -> None:
        # When the enum item is initialized, self._member_names_ contains the
        # names of the previously initialized items, so its length is the index
        # of this item.
        self.index = len(self._member_names_)

    # Bypass the slow Enum.__eq__
    __eq__ = object.__eq__

    # In Python 3, __hash__ must be defined if __eq__ is defined to stay
    # hashable.
    __hash__ = object.__hash__

    @classmethod
    def encode(
        cls,
        array: Union[
            EnumArray,
            numpy.int_,
            numpy.float_,
            numpy.object_,
        ],
    ) -> EnumArray:
        """
        Encode a string numpy array, an enum item numpy array, or an int numpy
        array into an :any:`EnumArray`. See :any:`EnumArray.decode` for
        decoding.

        :param numpy.ndarray array: Array of string identifiers, or of enum
                                    items, to encode.

        :returns: An :any:`EnumArray` encoding the input array values.
        :rtype: :any:`EnumArray`

        For instance:

        >>> string_identifier_array = asarray(['free_lodger', 'owner'])
        >>> encoded_array = HousingOccupancyStatus.encode(string_identifier_array)
        >>> encoded_array[0]
        2  # Encoded value

        >>> free_lodger = HousingOccupancyStatus.free_lodger
        >>> owner = HousingOccupancyStatus.owner
        >>> enum_item_array = asarray([free_lodger, owner])
        >>> encoded_array = HousingOccupancyStatus.encode(enum_item_array)
        >>> encoded_array[0]
        2  # Encoded value
        """
        if isinstance(array, EnumArray):
            return array

        # String array
        if isinstance(array, numpy.ndarray) and array.dtype.kind in {"U", "S"}:
            array = numpy.select(
                [array == item.name for item in cls],
                [item.index for item in cls],
            ).astype(ENUM_ARRAY_DTYPE)

        # Enum items arrays
        elif isinstance(array, numpy.ndarray) and array.dtype.kind == "O":
            # Ensure we are comparing the comparable. The problem this fixes:
            # On entering this method "cls" will generally come from
            # variable.possible_values, while the array values may come from
            # directly importing a module containing an Enum class. However,
            # variables (and hence their possible_values) are loaded by a call
            # to load_module, which gives them a different identity from the
            # ones imported in the usual way.
            #
            # So, instead of relying on the "cls" passed in, we use only its
            # name to check that the values in the array, if non-empty, are of
            # the right type.
            if len(array) > 0 and cls.__name__ is array[0].__class__.__name__:
                cls = array[0].__class__

            array = numpy.select(
                [array == item for item in cls],
                [item.index for item in cls],
            ).astype(ENUM_ARRAY_DTYPE)

        return EnumArray(array, cls)



import datetime

import numpy

from policyengine_core import enums
from policyengine_core.enums import Enum

VALUE_TYPES = {
    bool: {
        "dtype": numpy.bool_,
        "default": False,
        "json_type": "boolean",
        "formatted_value_type": "Boolean",
        "is_period_size_independent": True,
    },
    int: {
        "dtype": numpy.int32,
        "default": 0,
        "json_type": "integer",
        "formatted_value_type": "Int",
        "is_period_size_independent": False,
    },
    float: {
        "dtype": numpy.float32,
        "default": 0,
        "json_type": "number",
        "formatted_value_type": "Float",
        "is_period_size_independent": False,
    },
    str: {
        "dtype": object,
        "default": "",
        "json_type": "string",
        "formatted_value_type": "String",
        "is_period_size_independent": True,
    },
    Enum: {
        "dtype": enums.ENUM_ARRAY_DTYPE,
        "json_type": "string",
        "formatted_value_type": "String",
        "is_period_size_independent": True,
    },
    datetime.date: {
        "dtype": "datetime64[D]",
        "default": datetime.date.fromtimestamp(0),  # 0 == 1970-01-01
        "json_type": "string",
        "formatted_value_type": "Date",
        "is_period_size_independent": True,
    },
}


FORMULA_NAME_PREFIX = "formula"



"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a Person, a Household…

See https://openfisca.org/doc/key-concepts/variables.html
"""

from datetime import date

# Import from numpy the operations you need to apply on OpenFisca's population vectors
# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from numpy import where

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Person
from policyengine_core.periods import ETERNITY, MONTH
from policyengine_core.variables import Variable


# This variable is a pure input: it doesn't have a formula
class birth(Variable):
    value_type = date
    default_value = date(
        1970, 1, 1
    )  # By default, if no value is set for a simulation, we consider the people involved in a simulation to be born on the 1st of Jan 1970.
    entity = Person
    label = "Birth date"
    definition_period = ETERNITY  # This variable cannot change over time.
    reference = "https://en.wiktionary.org/wiki/birthdate"


class age(Variable):
    value_type = int
    entity = Person
    definition_period = MONTH
    label = "Person's age (in years)"

    def formula(person, period, _parameters):
        """
        Person's age (in years).

        A person's age is computed according to its birth date.
        """
        birth = person("birth", period)
        birth_year = birth.astype("datetime64[Y]").astype(int) + 1970
        birth_month = birth.astype("datetime64[M]").astype(int) % 12 + 1
        birth_day = (birth - birth.astype("datetime64[M]") + 1).astype(int)

        is_birthday_past = (birth_month < period.start.month) + (
            birth_month == period.start.month
        ) * (birth_day <= period.start.day)

        return (period.start.year - birth_year) - where(
            is_birthday_past, 0, 1
        )  # If the birthday is not passed this year, subtract one year



from .cycle_error import CycleError
from .empty_argument_error import EmptyArgumentError
from .nan_creation_error import NaNCreationError
from .parameter_not_found_error import ParameterNotFoundError
from .parameter_parsing_error import ParameterParsingError
from .period_mismatch_error import PeriodMismatchError
from .situation_parsing_error import SituationParsingError
from .spiral_error import SpiralError
from .variable_name_conflict_error import VariableNameConflictError
from .variable_not_found_error import VariableNotFoundError



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from policyengine_core.populations import GroupPopulation
from policyengine_core.projectors.projector import Projector


class FirstPersonToEntityProjector(Projector):
    """For instance famille.first_person."""

    def __init__(self, entity: "GroupPopulation", parent: Projector = None):
        self.target_entity = entity
        self.reference_entity = entity.members
        self.parent = parent

    def transform(self, result):
        return self.target_entity.value_from_first_person(result)



"""This file provides a function to load json example situations."""

import json
import os

DIR_PATH = os.path.dirname(os.path.abspath(__file__))


def parse(file_name):
    """Load json example situations."""
    file_path = os.path.join(DIR_PATH, file_name)
    with open(file_path, "r", encoding="utf8") as file:
        return json.loads(file.read())


single = parse("single.json")
couple = parse("couple.json")



import os


class VariableNotFoundError(Exception):
    """
    Exception raised when a variable has been queried but is not defined in the TaxBenefitSystem.
    """

    def __init__(self, variable_name, tax_benefit_system):
        """
        :param variable_name: Name of the variable that was queried.
        :param tax_benefit_system: Tax benefits system that does not contain `variable_name`
        """
        country_package_metadata = tax_benefit_system.get_package_metadata()
        country_package_name = country_package_metadata["name"]
        country_package_version = country_package_metadata["version"]
        if country_package_version:
            country_package_id = "{}@{}".format(
                country_package_name, country_package_version
            )
        else:
            country_package_id = country_package_name
        message = os.linesep.join(
            [
                "You tried to calculate or to set a value for variable '{0}', but it was not found in the loaded tax and benefit system ({1}).".format(
                    variable_name, country_package_id
                ),
                "Are you sure you spelled '{0}' correctly?".format(
                    variable_name
                ),
                "If this code used to work and suddenly does not, this is most probably linked to an update of the tax and benefit system.",
                "Look at its changelog to learn about renames and removals and update your code. If it is an official package,",
                "it is probably available on <https://github.com/openfisca/{0}/blob/master/CHANGELOG.md>.".format(
                    country_package_name
                ),
            ]
        )
        self.message = message
        self.variable_name = variable_name
        Exception.__init__(self, self.message)



"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a Person, a Household…

See https://openfisca.org/doc/key-concepts/variables.html
"""

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Household

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.periods import MONTH
from policyengine_core.variables import Variable


class total_benefits(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Sum of the benefits perceived by a household"
    reference = "https://stats.gov.example/benefits"

    def formula(household, period, _parameters):
        """Total benefits."""
        basic_income_i = household.members(
            "basic_income", period
        )  # Calculates the value of basic_income for each member of the household

        return +household.sum(
            basic_income_i
        ) + household(  # Sum the household members basic incomes
            "housing_allowance", period
        )


class total_taxes(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Sum of the taxes paid by a household"
    reference = "https://stats.gov.example/taxes"

    def formula(household, period, _parameters):
        """Total taxes."""
        income_tax_i = household.members("income_tax", period)
        social_security_contribution_i = household.members(
            "social_security_contribution", period
        )

        return (
            +household.sum(income_tax_i)
            + household.sum(social_security_contribution_i)
            + household("housing_tax", period.this_year) / 12
        )



import importlib
from argparse import ArgumentParser
from typing import List

import pandas as pd

from policyengine_core.data.dataset import Dataset
from policyengine_core.scripts import detect_country_package


def dataset_summary(datasets: List[Dataset]) -> str:
    years = list(sorted(list(set(sum([ds.years for ds in datasets], [])))))
    df = pd.DataFrame(
        {
            year: ["✓" if year in ds.years else "" for ds in datasets]
            for year in years
        },
        index=[ds.name for ds in datasets],
    )
    df = df.sort_values(by=list(df.columns[::-1]), ascending=False)
    return df.to_markdown(tablefmt="pretty")


def main(parser: ArgumentParser):
    # Get arguments as well as kwargs
    args, extra_args = parser.parse_known_args()

    country_package = args.country_package

    if country_package is None:
        country_package = detect_country_package()

    datasets: List[Dataset] = importlib.import_module(country_package).DATASETS

    if args.dataset == "datasets" and args.action == "list":
        print(dataset_summary(datasets))
        return
    dataset_by_name = {dataset.name: dataset for dataset in datasets}
    dataset = dataset_by_name[args.dataset]

    if args.action == "generate":
        dataset.generate(*extra_args)
    elif args.action == "download":
        dataset.download(*extra_args)
    elif args.action == "upload":
        dataset.upload(*extra_args)
    elif args.action == "remove":
        dataset.remove(*extra_args)
    elif args.action == "list":
        years = dataset.years
        if len(years) == 0:
            print("No saved datasets.")
        else:
            print("Saved datasets:")
            for year in years:
                filepath = dataset.file(year).absolute()
                print(
                    "  * " + filepath.name + "  | " + str(filepath.absolute())
                )
    else:
        raise ValueError(f"Action {args.action} not recognised.")



"""
This sub-package is used to define reforms.

A reform is a set of modifications to be applied to a reference tax and benefit system to carry out experiments.

See https://openfisca.org/doc/key-concepts/reforms.html
"""



from __future__ import annotations

import typing

import numpy

from policyengine_core.taxscales import AmountTaxScaleLike

if typing.TYPE_CHECKING:
    NumericalArray = typing.Union[numpy.int_, numpy.float_]


class MarginalAmountTaxScale(AmountTaxScaleLike):
    def calc(
        self,
        tax_base: NumericalArray,
        right: bool = False,
    ) -> numpy.float_:
        """
        Matches the input amount to a set of brackets and returns the sum of
        cell values from the lowest bracket to the one containing the input.
        """
        base1 = numpy.tile(tax_base, (len(self.thresholds), 1)).T

        thresholds1 = numpy.tile(
            numpy.hstack((self.thresholds, numpy.inf)), (len(tax_base), 1)
        )

        a = numpy.maximum(
            numpy.minimum(base1, thresholds1[:, 1:]) - thresholds1[:, :-1], 0
        )

        return numpy.dot(self.amounts, a.T > 0)



"""
This file defines a reform.

A reform is a set of modifications to be applied to a reference tax and benefit system to carry out experiments.

See https://openfisca.org/doc/key-concepts/reforms.html
"""

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.reforms import Reform


class removal_basic_income(Reform):
    def apply(self):
        """
        Apply reform.

        A reform always defines an `apply` method that builds the reformed tax and benefit system from the reference one.
        See https://openfisca.org/doc/coding-the-legislation/reforms.html#writing-a-reform

        Our reform neutralizes the `basic_income` variable. When this reform is applied, calculating `basic_income` will always return its default value, 0.
        """
        self.neutralize_variable("basic_income")



import logging
from typing import Any, Callable, Dict, List, Sequence, Type, TypeVar, Union

import numpy
import numpy as np
import pandas as pd
from numpy import logical_not as not_
from numpy import maximum as max_
from numpy import minimum as min_
from numpy import round as round_
from numpy import select, where

from policyengine_core.parameters.parameter_node import ParameterNode
from policyengine_core.periods.period_ import Period
from policyengine_core.populations.population import Population
from policyengine_core.types import ArrayLike, ArrayType
from policyengine_core.variables.variable import Variable

T = TypeVar("T")


def apply_thresholds(
    input: ArrayType[float],
    thresholds: ArrayLike[float],
    choices: ArrayLike[float],
) -> ArrayType[float]:
    """Makes a choice based on an input and thresholds.

    From a list of ``choices``, this function selects one of these values
    based on a list of inputs, depending on the value of each ``input`` within
    a list of ``thresholds``.

    Args:
        input: A list of inputs to make a choice from.
        thresholds: A list of thresholds to choose.
        choices: A list of the possible values to choose from.

    Returns:
        :obj:`numpy.ndarray` of :obj:`float`:
        A list of the values chosen.

    Raises:
        :exc:`AssertionError`: When the number of ``thresholds`` (t) and the
            number of choices (c) are not either t == c or t == c - 1.

    Examples:
        >>> input = numpy.array([4, 5, 6, 7, 8])
        >>> thresholds = [5, 7]
        >>> choices = [10, 15, 20]
        >>> apply_thresholds(input, thresholds, choices)
        array([10, 10, 15, 15, 20])

    """

    condlist: Sequence[ArrayType[bool]]
    condlist = [input <= threshold for threshold in thresholds]

    if len(condlist) == len(choices) - 1:
        # If a choice is provided for input > highest threshold, last condition
        # must be true to return it.
        condlist += [True]

    assert len(condlist) == len(choices), " ".join(
        [
            "'apply_thresholds' must be called with the same number of",
            "thresholds than choices, or one more choice.",
        ]
    )

    return numpy.select(condlist, choices)


def concat(this: ArrayLike[str], that: ArrayLike[str]) -> ArrayType[str]:
    """Concatenates the values of two arrays.

    Args:
        this: An array to concatenate.
        that: Another array to concatenate.

    Returns:
        :obj:`numpy.ndarray` of :obj:`float`:
        An array with the concatenated values.

    Examples:
        >>> this = ["this", "that"]
        >>> that = numpy.array([1, 2.5])
        >>> concat(this, that)
        array(['this1.0', 'that2.5']...)

    """

    if isinstance(this, numpy.ndarray) and not numpy.issubdtype(
        this.dtype, numpy.str_
    ):

        this = this.astype("str")

    if isinstance(that, numpy.ndarray) and not numpy.issubdtype(
        that.dtype, numpy.str_
    ):

        that = that.astype("str")

    return numpy.char.add(this, that)


def switch(
    conditions: ArrayType[Any],
    value_by_condition: Dict[float, T],
) -> ArrayType[T]:
    """Mimicks a switch statement.

    Given an array of conditions, returns an array of the same size,
    replacing each condition item with the matching given value.

    Args:
        conditions: An array of conditions.
        value_by_condition: Values to replace for each condition.

    Returns:
        :obj:`numpy.ndarray`:
        An array with the replaced values.

    Raises:
        :exc:`AssertionError`: When ``value_by_condition`` is empty.

    Examples:
        >>> conditions = numpy.array([1, 1, 1, 2])
        >>> value_by_condition = {1: 80, 2: 90}
        >>> switch(conditions, value_by_condition)
        array([80, 80, 80, 90])

    """

    assert (
        len(value_by_condition) > 0
    ), "'switch' must be called with at least one value."

    condlist = [
        conditions == condition for condition in value_by_condition.keys()
    ]

    return numpy.select(condlist, value_by_condition.values())


def for_each_variable(
    entity: Population,
    period: Period,
    variables: List[str],
    agg_func: str = "add",
    group_agg_func: str = "add",
    options: List[str] = None,
) -> ArrayLike:
    """Applies operations to lists of variables.

    Args:
        entity (Population): The entity population, as passed in formulas.
        period (Period): The period, as pass in formulas.
        variables (List[str]): A list of variable names.
        agg_func (str, optional): The operation to apply to combine variable results. Defaults to "add".
        group_agg_func (str, optional): The operation to apply to transform values to the target entity level. Defaults to "add".
        options (List[str], optional): Options to pass to the `entity(variable, period)` call. Defaults to None.

    Raises:
        ValueError: If any target variable is not at or below the target entity level.

    Returns:
        ArrayLike: The result of the operation.
    """
    result = None
    agg_func = dict(
        add=lambda x, y: x + y, multiply=lambda x, y: x * y, max=max_, min=min_
    )[agg_func]
    if not entity.entity.is_person:
        group_agg_func = dict(
            add=entity.sum, all=entity.all, max=entity.max, min=entity.min
        )[group_agg_func]
    for variable in variables:
        variable_entity = entity.entity.get_variable(variable).entity
        if variable_entity.key == entity.entity.key:
            values = entity(variable, period, options=options)
        elif variable_entity.is_person:
            values = group_agg_func(
                entity.members(variable, period, options=options)
            )
        elif entity.entity.is_person:
            raise ValueError(
                f"You requested to aggregate {variable} (defined for {variable_entity.plural}) to {entity.entity.plural}, but this is not yet implemented."
            )
        else:  # Group-to-group aggregation
            variable_population = entity.simulation.populations[
                variable_entity.key
            ]
            person_shares = variable_population.project(
                variable_population(variable, period)
            ) / variable_population.project(variable_population.nb_persons())
            values = entity.sum(person_shares)
        if result is None:
            result = values
        else:
            result = agg_func(result, values)
    return result


def add(
    entity: Population,
    period: Period,
    variables: List[str],
    options: List[str] = None,
):
    """Sums a list of variables.

    Args:
        entity (Population): The entity population, as passed in formulas.
        period (Period): The period, as pass in formulas.
        variables (List[str]): A list of variable names.
        options (List[str], optional): Options to pass to the `entity(variable, period)` call. Defaults to None.

    Raises:
        ValueError: If any target variable is not at or below the target entity level.

    Returns:
        ArrayLike: The result of the operation.
    """
    return for_each_variable(
        entity, period, variables, agg_func="add", options=options
    )


def aggr(entity, period, variables, options=None):
    """Sums a list of variables belonging to entity members.

    Args:
        entity (Population): The entity population, as passed in formulas.
        period (Period): The period, as pass in formulas.
        variables (List[str]): A list of variable names.
        options (List[str], optional): Options to pass to the `entity(variable, period)` call. Defaults to None.

    Raises:
        ValueError: If any target variable is not below the target entity level.

    Returns:
        ArrayLike: The result of the operation.
    """
    return for_each_variable(
        entity,
        period,
        variables,
        agg_func="add",
        group_agg_func="add",
        options=options,
    )


def and_(
    entity: Population,
    period: Period,
    variables: List[str],
    options: List[str] = None,
):
    """Performs a logical and operation on a list of variables.

    Args:
        entity (Population): The entity population, as passed in formulas.
        period (Period): The period, as pass in formulas.
        variables (List[str]): A list of variable names.
        options (List[str], optional): Options to pass to the `entity(variable, period)` call. Defaults to None.

    Raises:
        ValueError: If any target variable is not at the target entity level.

    Returns:
        ArrayLike: The result of the operation.
    """
    return for_each_variable(
        entity, period, variables, agg_func="multiply", options=options
    )


or_ = add
any_ = or_
multiply = and_

select = np.select


clip = np.clip
inf = np.inf

WEEKS_IN_YEAR = 52
MONTHS_IN_YEAR = 12


def amount_over(amount: ArrayLike, threshold: float) -> ArrayLike:
    """Calculates the amounts over a threshold.

    Args:
        amount (ArrayLike): The amount to calculate for.
        threshold_1 (float): The threshold.

    Returns:
        ArrayLike: The amounts over the threshold.
    """
    logging.debug(
        "amount_over(x, y) is deprecated, use max_(x - y, 0) instead."
    )
    return max_(0, amount - threshold)


def amount_between(
    amount: ArrayLike, threshold_1: float, threshold_2: float
) -> ArrayLike:
    """Calculates the amounts between two thresholds.

    Args:
        amount (ArrayLike): The amount to calculate for.
        threshold_1 (float): The lower threshold.
        threshold_2 (float): The upper threshold.

    Returns:
        ArrayLike: The amounts between the thresholds.
    """
    return clip(amount, threshold_1, threshold_2) - threshold_1


def random(entity, reset=True):
    if reset:
        np.random.seed(0)
    x = np.random.rand(entity.count)
    return x


def is_in(values: ArrayLike, *targets: list) -> ArrayLike:
    """Returns true if the value is in the list of targets.

    Args:
        values (ArrayLike): The values to test.

    Returns:
        ArrayLike: True if the value is in the list of targets.
    """
    if (len(targets) == 1) and isinstance(targets[0], list):
        targets = targets[0]
    return np.any([values == target for target in targets], axis=0)


def between(
    values: ArrayLike, lower: float, upper: float, inclusive: str = "both"
) -> ArrayLike:
    """Returns true if values are between lower and upper.

    Args:
        values (ArrayLike): The input array.
        lower (float): The lower bound.
        upper (float): The upper bound.
        inclusive (bool, optional): Whether to include or exclude the bounds. Defaults to True.

    Returns:
        ArrayLike: The resulting array.
    """
    return pd.Series(values).between(lower, upper, inclusive=inclusive)


def uprated(by: str = None, start_year: int = 2015) -> Callable:
    """Attaches a formula applying an uprating factor to input variables (going back as far as 2015).

    Args:
        by (str, optional): The name of the parameter (under parameters.uprating). Defaults to None (no uprating applied).

    Returns:
        Callable: A class decorator.
    """

    def uprater(variable: Type[Variable]) -> type:
        if hasattr(variable, f"formula_{start_year}"):
            return variable

        formula = variable.formula if hasattr(variable, "formula") else None

        variable.metadata = {
            "uprating": by,
        }

        def formula_start_year(entity, period, parameters):
            if by is None:
                return entity(variable.__name__, period.last_year)
            else:
                current_parameter = parameters(period)
                last_year_parameter = parameters(period.last_year)
                for name in by.split("."):
                    current_parameter = getattr(current_parameter, name)
                    last_year_parameter = getattr(last_year_parameter, name)
                uprating = current_parameter / last_year_parameter
                old = entity(variable.__name__, period.last_year)
                if (formula is not None) and (all(old) == 0):
                    # If no values have been inputted, don't uprate and
                    # instead use the previous formula on the current period.
                    return formula(entity, period, parameters)
                return uprating * old

        formula_start_year.__name__ = f"formula_{start_year}"
        setattr(variable, formula_start_year.__name__, formula_start_year)
        return variable

    return uprater


def carried_over(variable: type) -> type:
    return uprated()(variable)


def sum_of_variables(variables: Union[List[str], str]) -> Callable:
    """Returns a function that sums the values of a list of variables.

    Args:
        variables (Union[List[str], str]): A list of variable names.

    Returns:
        Callable: A function that sums the values of the variables.
    """

    def sum_of_variables(entity, period, parameters):
        if isinstance(variables, str):
            # A string parameter name is passed
            node = parameters(period)
            for name in variables.split("."):
                node = getattr(node, name)
            variable_names = node
        else:
            variable_names = variables
        return add(entity, period, variable_names)

    return sum_of_variables


any_of_variables = sum_of_variables


def index_(
    into: ParameterNode,
    indices: Union[ArrayLike, List[ArrayLike]],
    where: ArrayLike,
    fill: float = 0,
) -> ArrayLike:
    """Indexes into a object, but only when a condition is true. This improves
    performance over `np.where`, which will index all values and then filter the result.

    Args:
        into (Parameter): The parameter to index into.
        indices (Union[ArrayLike, List[ArrayLike]]): The full, un-filtered index array. Can be a list of arrays
            for sequential indexing.
        where (ArrayLike): A filter for indexing.
        fill (float, optional): The value to fill where `index_where` is False. Defaults to 0.

    Returns:
        ArrayLike: The indexed result.
    """
    if where.sum() == 0:
        return np.ones(where.shape) * fill

    if isinstance(indices, list):
        result = np.empty_like(indices[0])
        intermediate_result = into
        for i in range(len(indices)):
            intermediate_result = intermediate_result[indices[i][where]]
        result[where] = intermediate_result
    else:
        result = np.empty_like(indices)
        result[where] = into[indices[where]]
    result[~where] = fill
    return result.astype(float)



from policyengine_core.country_template.constants import COUNTRY_DIR
from policyengine_core.data import Dataset
from policyengine_core.periods import ETERNITY, MONTH, period


class CountryTemplateDataset(Dataset):
    # Specify metadata used to describe and store the dataset.
    name = "country_template_dataset"
    label = "Country template dataset"
    folder_path = COUNTRY_DIR / "data" / "storage"
    data_format = Dataset.TIME_PERIOD_ARRAYS

    # The generation function is the most important part: it defines
    # how the dataset is generated from the raw data for a given year.
    def generate(self, year: int) -> None:
        person_id = [0, 1, 2]
        household_id = [0, 1]
        person_household_id = [0, 0, 1]
        person_household_role = ["parent", "child", "parent"]
        salary = [100, 0, 200]
        salary_time_period = period("2022-01")
        weight = [1e6, 1.2e6]
        weight_time_period = period("2022")
        data = {
            "person_id": {ETERNITY: person_id},
            "household_id": {ETERNITY: household_id},
            "person_household_id": {ETERNITY: person_household_id},
            "person_household_role": {ETERNITY: person_household_role},
            "salary": {salary_time_period: salary},
            "household_weight": {weight_time_period: weight},
        }
        self.save_variable_values(year, data)


# Important: we must instantiate datasets. This tests their validity and adds dynamic logic.
CountryTemplateDataset = CountryTemplateDataset()



from policyengine_core.parameters.parameter import Parameter
from policyengine_core.parameters.parameter_at_instant import (
    ParameterAtInstant,
)
from policyengine_core.parameters.parameter_node import ParameterNode
from policyengine_core.periods import instant


def interpolate_parameters(root: ParameterNode) -> ParameterNode:
    """Interpolates parameters according to their metadata.

    Args:
        root (ParameterNode): The root of the parameter tree.

    Returns:
        ParameterNode: The same root, with interpolation applied to descendants.
    """
    for parameter in root.get_descendants():
        if isinstance(parameter, Parameter):
            if (
                "interpolation" in parameter.metadata
                and not parameter.metadata["interpolation"].get(
                    "completed", False
                )
            ):
                interpolated_entries = []
                for i in range(len(parameter.values_list) - 1):
                    # For each gap in parameter values
                    start = instant(parameter.values_list[::-1][i].instant_str)
                    num_intervals = 1
                    # Find the number of intervals to fill
                    interval_size = parameter.metadata["interpolation"][
                        "interval"
                    ]
                    parameter_dates = [
                        at_instant.instant_str
                        for at_instant in parameter.values_list
                    ]
                    while (
                        str(start.offset(num_intervals, interval_size))
                        < parameter_dates[0]
                    ):
                        num_intervals += 1
                    # Interpolate in each interval
                    for j in range(1, num_intervals):
                        start_str = str(
                            start.offset(
                                j,
                                parameter.metadata["interpolation"][
                                    "interval"
                                ],
                            )
                        )
                        start_value = parameter.values_list[::-1][i].value
                        end_value = parameter.values_list[::-1][i + 1].value
                        new_value = (
                            start_value
                            + (end_value - start_value) * j / num_intervals
                        )
                        interpolated_entries += [
                            ParameterAtInstant(
                                parameter.name, start_str, data=new_value
                            )
                        ]
                for entry in interpolated_entries:
                    parameter.values_list.append(entry)
                parameter.values_list.sort(
                    key=lambda x: x.instant_str, reverse=True
                )
                parameter.metadata["interpolation"]["completed"] = True
    return root



import importlib
import logging
import os
import re
from pathlib import Path
from typing import Callable, Dict, Union

import h5py
import numpy as np
import pandas as pd


class Dataset:
    """The `Dataset` class is a base class for datasets used directly or indirectly for OpenFisca models.
    A dataset defines a generation function to create it from other data, and this class provides common features
    like cloud storage, metadata and loading."""

    name: str = None
    """The name of the dataset. This is used to generate filenames and is used as the key in the `datasets` dictionary."""
    label: str = None
    """The label of the dataset. This is used for logging and is used as the key in the `datasets` dictionary."""
    is_openfisca_compatible: bool = True
    """Whether the dataset is compatible with OpenFisca. If True, the dataset will be stored as a collection of arrays. If False, the dataset will be stored as a collection of tables."""
    data_format: str = None
    """The format of the dataset. This can be either `Dataset.ARRAYS`, `Dataset.TIME_PERIOD_ARRAYS` or `Dataset.TABLES`. If `Dataset.ARRAYS`, the dataset is stored as a collection of arrays. If `Dataset.TIME_PERIOD_ARRAYS`, the dataset is stored as a collection of arrays, with one array per time period. If `Dataset.TABLES`, the dataset is stored as a collection of tables (DataFrames)."""
    folder_path: str = None
    """The path to the folder where the dataset is stored (in .h5 files)."""

    # Data formats
    TABLES = "tables"
    ARRAYS = "arrays"
    TIME_PERIOD_ARRAYS = "time_period_arrays"

    def __init__(self):
        # Setup dataset
        if self.folder_path is None:
            raise ValueError(
                "Dataset folder_path must be specified in the dataset class definition."
            )

        if not self.folder_path.exists():
            logging.info(
                f"Creating folder {self.folder_path} for {self.label} dataset."
            )
            self.folder_path.mkdir(parents=True)

        assert (
            self.name
        ), "You tried to instantiate a Dataset object, but no name has been provided."
        assert (
            self.label
        ), "You tried to instantiate a Dataset object, but no label has been provided."

        if self.data_format is None:
            if not self.is_openfisca_compatible:
                # Assume that external (raw) datasets are a collection of tables.
                self.data_format = Dataset.TABLES
            else:
                # OpenFisca datasets are a collection of arrays.
                self.data_format = Dataset.ARRAYS

        assert self.data_format in [
            Dataset.TABLES,
            Dataset.ARRAYS,
            Dataset.TIME_PERIOD_ARRAYS,
        ], "You tried to instantiate a Dataset object, but your data_format attribute is invalid."

        # Ensure typed arguments are enforced in `generate`

        def cast_first_arg_as_int(fn: Callable) -> Callable:
            def wrapper(year: str, *args, **kwargs):
                year = int(year)
                return fn(year, *args, **kwargs)

            return wrapper

        self.generate = cast_first_arg_as_int(self.generate)
        self.download = cast_first_arg_as_int(self.download)
        self.upload = cast_first_arg_as_int(self.upload)

    def filename(self, year: int) -> str:
        """Returns the filename of the dataset for a given year.

        Args:
            year (int): The year of the dataset to generate.

        Returns:
            str: The filename of the dataset.
        """
        return f"{self.name}_{year}.h5"

    def file(self, year: int) -> Path:
        """Returns the path to the dataset for a given year.

        Args:
            year (int): The year of the dataset to generate.

        Returns:
            Path: The path to the dataset.
        """
        return self.folder_path / self.filename(year)

    def load(
        self, year: int, key: str = None, mode: str = "r"
    ) -> Union[h5py.File, np.array, pd.DataFrame, pd.HDFStore]:
        """Loads the dataset for a given year, returning a H5 file reader. You can then access the
        dataset like a dictionary (e.g.e Dataset.load(2022)["variable"]).

        Args:
            year (int): The year of the dataset to load.
            key (str, optional): The key to load. Defaults to None.
            mode (str, optional): The mode to open the file with. Defaults to "r".

        Returns:
            Union[h5py.File, np.array, pd.DataFrame, pd.HDFStore]: The dataset.
        """
        file = self.folder_path / self.filename(year)
        if self.data_format in (Dataset.ARRAYS, Dataset.TIME_PERIOD_ARRAYS):
            if key is None:
                # If no key provided, return the basic H5 reader.
                return h5py.File(file, mode=mode)
            else:
                # If key provided, return only the values requested.
                with h5py.File(file, mode=mode) as f:
                    values = np.array(f[key])
                return values
        elif self.data_format == Dataset.TABLES:
            if key is None:
                # Non-openfisca datasets are assumed to be of the format (table name: [table], ...).
                return pd.HDFStore(file)
            else:
                # If a table name is provided, return that table.
                with pd.HDFStore(file) as f:
                    values = f[key]
                return values
        else:
            raise ValueError(
                f"Invalid data format {self.data_format} for dataset {self.label}."
            )

    def save(self, year: int, key: str, values: Union[np.array, pd.DataFrame]):
        """Overwrites the values for `key` with `values`.

        Args:
            year (int): The year of the dataset to save.
            key (str): The key to save.
            values (Union[np.array, pd.DataFrame]): The values to save.
        """
        file = self.folder_path / self.filename(year)
        if self.data_format in (Dataset.ARRAYS, Dataset.TIME_PERIOD_ARRAYS):
            with h5py.File(file, "a") as f:
                # Overwrite if existing
                if key in f:
                    del f[key]
                f.create_dataset(key, data=values)
        elif self.data_format == Dataset.TABLES:
            with pd.HDFStore(file, "a") as f:
                f.put(key, values)
        else:
            raise ValueError(
                f"Invalid data format {self.data_format} for dataset {self.label}."
            )

    def save_variable_values(
        self, year: int, data: Dict[str, Dict[str, np.array]]
    ) -> None:
        """Writes data values for variables in specific time periods.

        Args:
            year (int): The year of the dataset to save.
            data (Dict[str, Dict[str, np.array]]): The data to save. Should be nested in the form {variable_name: {time_period: values}}.
        """
        if self.data_format is not self.TIME_PERIOD_ARRAYS:
            raise ValueError(
                f"Invalid data format {self.data_format} for dataset {self.label}."
            )

        file = self.folder_path / self.filename(year)
        with h5py.File(file, "a") as f:
            for variable, values in data.items():
                for time_period, value in values.items():
                    key = f"{variable}/{time_period}"
                    # Overwrite if existing
                    if key in f:
                        del f[key]
                    f.create_dataset(key, data=value)

    def keys(self, year: int):
        """Returns the keys of the dataset for a given year.

        Args:
            year (int): The year of the dataset to generate.

        Returns:
            list: The keys of the dataset.
        """
        if self.data_format in (Dataset.ARRAYS, Dataset.TIME_PERIOD_ARRAYS):
            with h5py.File(self.file(year), mode="r") as f:
                return list(f.keys())
        elif self.data_format == Dataset.TABLES:
            with pd.HDFStore(self.file(year)) as f:
                return list(f.keys())
        else:
            raise ValueError(
                f"Invalid data format {self.data_format} for dataset {self.label}."
            )

    def generate(self, year: int):
        """Generates the dataset for a given year (all datasets should implement this method).

        Args:
            year (int): The year of the dataset to generate.

        Raises:
            NotImplementedError: If the function has not been overriden.
        """
        raise NotImplementedError(
            f"You tried to generate the dataset for year {year} of {self.label}, but no dataset generation implementation has been provided for {self.label}."
        )

    def download(self, year: int):
        """Downloads the dataset for a given year (all datasets should implement this method).

        Args:
            year (int): The year of the dataset to download.

        Raises:
            NotImplementedError: If the function has not been overriden.
        """
        raise NotImplementedError(
            f"You tried to download the dataset for year {year} of {self.label}, but no dataset download implementation has been provided for {self.label}."
        )

    def upload(self, year: int):
        """Uploads the dataset for a given year (all datasets should implement this method).

        Args:
            year (int): The year of the dataset to upload.

        Raises:
            NotImplementedError: If the function has not been overriden.
        """
        raise NotImplementedError(
            f"You tried to upload the dataset for year {year} of {self.label}, but no dataset upload implementation has been provided for {self.label}."
        )

    def remove(self, year: int = None):
        """Removes the dataset file for a given year.

        Args:
            year (int, optional): The year of the dataset to remove. Defaults to all.
        """
        if year is None:
            filenames = map(self.filename(year), self.years)
        else:
            filenames = (self.filename(year),)
        for filename in filenames:
            filepath = self.folder_path / filename
            if filepath.exists():
                os.remove(filepath)

    def remove_all(self):
        """Removes all dataset files."""
        for year in self.years:
            self.remove(year)

    @property
    def years(self):
        """Returns the years for which the dataset has been generated."""
        pattern = re.compile(f"\n{self.name}_([0-9]+).h5")
        return list(
            map(
                int,
                pattern.findall(
                    "\n"
                    + "\n".join(
                        map(lambda path: path.name, self.folder_path.iterdir())
                    )
                ),
            )
        )



import warnings
from typing import List

from policyengine_core.warnings import MemoryConfigWarning


class MemoryConfig:
    def __init__(
        self,
        max_memory_occupation: float,
        priority_variables: List[str] = None,
        variables_to_drop: List[str] = None,
    ):
        message = [
            "Memory configuration is a feature that is still currently under experimentation.",
            "You are very welcome to use it and send us precious feedback,",
            "but keep in mind that the way it is used might change without any major version bump.",
        ]
        warnings.warn(" ".join(message), MemoryConfigWarning)

        self.max_memory_occupation = float(max_memory_occupation)
        if self.max_memory_occupation > 1:
            raise ValueError("max_memory_occupation must be <= 1")
        self.max_memory_occupation_pc = self.max_memory_occupation * 100
        self.priority_variables = (
            set(priority_variables) if priority_variables else set()
        )
        self.variables_to_drop = (
            set(variables_to_drop) if variables_to_drop else set()
        )



from typing import List, Union

from .reference import Reference


class UpratingIndex(str):
    """
    The name of a parameter whose values are used to uprate a parameter node.
    """


class SelfUprating(UpratingIndex):
    """
    A special value ("self") that indicates that the parameter node should be uprated based on the fixed rate increase between two points in its history.
    """


class UpratingRoundingConfig(dict):
    type: str
    """Either 'nearest', 'upwards' or 'downwards'."""

    interval: float
    """The interval to round to."""


class UpratingSchema(dict):
    """
    A schema for uprating a parameter node.
    """

    parameter: UpratingIndex
    """
    The name of the parameter whose values are used to uprate the parameter node.
    """

    start_instant: int
    """
    The instant to add uprated values after.
    """

    type: UpratingRoundingConfig
    """
    The rounding configuration to use when uprating.
    """


class ParameterMetadata:
    """A single legislative value that can change over time."""

    name: str
    """The name identifying this parameter. Should be snake-case and Python-safe."""

    label: str
    """The display text to use for the parameter. Should be short and less than a full sentence."""

    description: str
    """A longer description of the parameter."""

    documentation: str
    """Any relevant information about the parameter's *implementation* (not its legislative meaning- use the description for that field)."""

    reference: List[Reference]
    """A list of references to legislation or other documents that define this parameter."""

    unit: str
    """The real-world meaning of a particular value."""

    uprating: Union[UpratingIndex, UpratingSchema]
    """The schema for uprating this parameter.

    The process for uprating parameter $X$ based on parameter $Y$ is as follows:
    
    1. Look at $X$ and $Y$ at the same time, and move forward in history.

    2. If $Y$ increases by $z$%, then increase $X$ by $z$%.
    """



from __future__ import annotations

import typing
from typing import Union

import numpy

from policyengine_core import parameters

from .. import tracers

ParameterNode = Union[
    parameters.VectorialParameterNodeAtInstant,
    parameters.ParameterNodeAtInstant,
]

if typing.TYPE_CHECKING:
    from numpy.typing import ArrayLike

    Child = Union[ParameterNode, ArrayLike]


class TracingParameterNodeAtInstant:
    def __init__(
        self,
        parameter_node_at_instant: ParameterNode,
        tracer: tracers.FullTracer,
        branch_name: str,
    ) -> None:
        self.parameter_node_at_instant = parameter_node_at_instant
        self.tracer = tracer
        self.branch_name = branch_name

    def __getattr__(
        self,
        key: str,
    ) -> Union[TracingParameterNodeAtInstant, Child]:
        child = getattr(self.parameter_node_at_instant, key)
        return self.get_traced_child(child, key)

    def __getitem__(
        self,
        key: Union[str, ArrayLike],
    ) -> Union[TracingParameterNodeAtInstant, Child]:
        child = self.parameter_node_at_instant[key]
        return self.get_traced_child(child, key)

    def get_traced_child(
        self,
        child: Child,
        key: Union[str, ArrayLike],
    ) -> Union[TracingParameterNodeAtInstant, Child]:
        period = self.parameter_node_at_instant._instant_str

        if isinstance(
            child,
            (
                parameters.ParameterNodeAtInstant,
                parameters.VectorialParameterNodeAtInstant,
            ),
        ):
            return TracingParameterNodeAtInstant(
                child, self.tracer, self.branch_name
            )

        if not isinstance(key, str) or isinstance(
            self.parameter_node_at_instant,
            parameters.VectorialParameterNodeAtInstant,
        ):
            # In case of vectorization, we keep the parent node name as, for
            # instance, rate[status].zone1 is best described as the value of
            # "rate".
            name = self.parameter_node_at_instant._name

        else:
            name = ".".join([self.parameter_node_at_instant._name, key])

        if isinstance(
            child, (numpy.ndarray,) + parameters.ALLOWED_PARAM_TYPES
        ):
            self.tracer.record_parameter_access(
                name, period, self.branch_name, child
            )

        return child



from __future__ import annotations

import logging
import typing

import numpy

from policyengine_core.taxscales.marginal_rate_tax_scale import (
    MarginalRateTaxScale,
)
from policyengine_core.taxscales.rate_tax_scale_like import RateTaxScaleLike

log = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    NumericalArray = typing.Union[numpy.int_, numpy.float_]


class LinearAverageRateTaxScale(RateTaxScaleLike):
    def calc(
        self,
        tax_base: NumericalArray,
        right: bool = False,
    ) -> numpy.float_:
        if len(self.rates) == 1:
            return tax_base * self.rates[0]

        tiled_base = numpy.tile(tax_base, (len(self.thresholds) - 1, 1)).T
        tiled_thresholds = numpy.tile(self.thresholds, (len(tax_base), 1))

        bracket_dummy = (tiled_base >= tiled_thresholds[:, :-1]) * (
            +tiled_base < tiled_thresholds[:, 1:]
        )

        rates_array = numpy.array(self.rates)
        thresholds_array = numpy.array(self.thresholds)

        rate_slope = (rates_array[1:] - rates_array[:-1]) / (
            +thresholds_array[1:] - thresholds_array[:-1]
        )

        average_rate_slope = numpy.dot(bracket_dummy, rate_slope.T)

        bracket_average_start_rate = numpy.dot(bracket_dummy, rates_array[:-1])
        bracket_threshold = numpy.dot(bracket_dummy, thresholds_array[:-1])

        log.info(f"bracket_average_start_rate :  {bracket_average_start_rate}")
        log.info(f"average_rate_slope:  {average_rate_slope}")

        return tax_base * (
            +bracket_average_start_rate
            + (tax_base - bracket_threshold) * average_rate_slope
        )

    def to_marginal(self) -> MarginalRateTaxScale:
        marginal_tax_scale = MarginalRateTaxScale(
            name=self.name,
            option=self.option,
            unit=self.unit,
        )

        previous_i = 0
        previous_threshold = 0

        for threshold, rate in zip(self.thresholds[1:], self.rates[1:]):
            if threshold != float("Inf"):
                i = rate * threshold
                marginal_tax_scale.add_bracket(
                    previous_threshold,
                    (i - previous_i) / (threshold - previous_threshold),
                )
                previous_i = i
                previous_threshold = threshold

        marginal_tax_scale.add_bracket(previous_threshold, rate)

        return marginal_tax_scale



from policyengine_core.errors import (
    ParameterNotFoundError,
    ParameterParsingError,
)

from .at_instant_like import AtInstantLike
from .config import (
    ALLOWED_PARAM_TYPES,
    COMMON_KEYS,
    FILE_EXTENSIONS,
    date_constructor,
    dict_no_duplicate_constructor,
)
from .helpers import contains_nan, load_parameter_file
from .operations import (
    homogenize_parameter_structures,
    interpolate_parameters,
    propagate_parameter_metadata,
    uprate_parameters,
)
from .parameter import Parameter
from .parameter_at_instant import ParameterAtInstant
from .parameter_node import ParameterNode
from .parameter_node_at_instant import ParameterNodeAtInstant
from .parameter_scale import ParameterScale
from .parameter_scale_bracket import ParameterScaleBracket
from .vectorial_parameter_node_at_instant import (
    VectorialParameterNodeAtInstant,
)



import copy
import os
import typing
from typing import Iterable, List, Type, Union

from policyengine_core import commons, parameters, tools
from policyengine_core.data_structures import Reference
from policyengine_core.periods.instant_ import Instant

from .at_instant_like import AtInstantLike
from .parameter import Parameter
from .parameter_node_at_instant import ParameterNodeAtInstant
from .config import COMMON_KEYS, FILE_EXTENSIONS
from .helpers import (
    load_parameter_file,
    _compose_name,
    _validate_parameter,
    _parse_child,
    _load_yaml_file,
)

EXCLUDED_PARAMETER_CHILD_NAMES = ["reference", "__pycache__"]


class ParameterNode(AtInstantLike):
    """
    A node in the legislation `parameter tree <https://openfisca.org/doc/coding-the-legislation/legislation_parameters.html>`_.
    """

    _allowed_keys: typing.Optional[
        typing.Iterable[str]
    ] = None  # By default, no restriction on the keys

    parent: "ParameterNode" = None
    """The parent of the node, or None if the node is the root of the tree."""

    def __init__(
        self,
        name: str = "",
        directory_path: str = None,
        data: dict = None,
        file_path: str = None,
    ):
        """
        Instantiate a ParameterNode either from a dict, (using `data`), or from a directory containing YAML files (using `directory_path`).

        :param str name: Name of the node, eg "taxes.some_tax".
        :param str directory_path: Directory containing YAML files describing the node.
        :param dict data: Object representing the parameter node. It usually has been extracted from a YAML file.
        :param str file_path: YAML file from which the `data` has been extracted from.


        Instantiate a ParameterNode from a dict:

        >>> node = ParameterNode('basic_income', data = {
            'amount': {
              'values': {
                "2015-01-01": {'value': 550},
                "2016-01-01": {'value': 600}
                }
              },
            'min_age': {
              'values': {
                "2015-01-01": {'value': 25},
                "2016-01-01": {'value': 18}
                }
              },
            })

        Instantiate a ParameterNode from a directory containing YAML parameter files:

        >>> node = ParameterNode('benefits', directory_path = '/path/to/country_package/parameters/benefits')
        """
        self.name: str = name
        self.children: typing.Dict[
            str,
            typing.Union[ParameterNode, Parameter, parameters.ParameterScale],
        ] = {}
        self.description: str = None
        self.documentation: str = None
        self.file_path: str = None
        self.metadata: dict = {}

        if directory_path:
            self.file_path = directory_path
            for child_name in os.listdir(directory_path):
                child_path = os.path.join(directory_path, child_name)
                if os.path.isfile(child_path):
                    child_name, ext = os.path.splitext(child_name)

                    # We ignore non-YAML files
                    if ext not in FILE_EXTENSIONS:
                        continue

                    if child_name == "index":
                        data = _load_yaml_file(child_path) or {}
                        _validate_parameter(
                            self, data, allowed_keys=COMMON_KEYS
                        )
                        self.description = data.get("description")
                        self.documentation = data.get("documentation")
                        self.metadata.update(data.get("metadata", {}))
                    elif child_name not in EXCLUDED_PARAMETER_CHILD_NAMES:
                        child_name_expanded = _compose_name(name, child_name)
                        child = load_parameter_file(
                            child_path, child_name_expanded
                        )
                        self.add_child(child_name, child)

                elif os.path.isdir(child_path):
                    child_name = os.path.basename(child_path)
                    child_name_expanded = _compose_name(name, child_name)
                    child = ParameterNode(
                        child_name_expanded, directory_path=child_path
                    )
                    self.add_child(child_name, child)

        else:
            self.file_path = file_path
            _validate_parameter(
                self, data, data_type=dict, allowed_keys=self._allowed_keys
            )
            self.description = data.get("description")
            self.documentation = data.get("documentation")
            self.metadata.update(data.get("metadata", {}))
            for child_name, child in data.items():
                if (
                    child_name in COMMON_KEYS
                    or child_name in EXCLUDED_PARAMETER_CHILD_NAMES
                ):
                    continue  # do not treat reserved keys as subparameters.

                child_name = str(child_name)
                child_name_expanded = _compose_name(name, child_name)
                child = _parse_child(child_name_expanded, child, file_path)
                self.add_child(child_name, child)

    def merge(self, other: "ParameterNode") -> None:
        """
        Merges another ParameterNode into the current node.

        In case of child name conflict, the other node child will replace the current node child.
        """
        for child_name, child in other.children.items():
            self.add_child(child_name, child)

    def add_child(self, name: str, child: Union["ParameterNode", Parameter]):
        """
        Add a new child to the node.

        :param name: Name of the child that must be used to access that child. Should not contain anything that could interfere with the operator `.` (dot).
        :param child: The new child, an instance of :class:`.ParameterScale` or :class:`.Parameter` or :class:`.ParameterNode`.
        """
        if name in self.children:
            raise ValueError(
                "{} has already a child named {}".format(self.name, name)
            )
        if not (
            isinstance(child, ParameterNode)
            or isinstance(child, Parameter)
            or isinstance(child, parameters.ParameterScale)
        ):
            raise TypeError(
                "child must be of type ParameterNode, Parameter, or Scale. Instead got {}".format(
                    type(child)
                )
            )
        self.children[name] = child
        setattr(self, name, child)
        child.parent = self

    def __repr__(self) -> str:
        result = os.linesep.join(
            [
                os.linesep.join(["{}:", "{}"]).format(
                    name, tools.indent(repr(value))
                )
                for name, value in sorted(self.children.items())
            ]
        )
        return result

    def get_descendants(self) -> Iterable[Union["ParameterNode", Parameter]]:
        """
        Return a generator containing all the parameters and nodes recursively contained in this `ParameterNode`
        """
        for child in self.children.values():
            yield child
            yield from child.get_descendants()

    def clone(self) -> "ParameterNode":
        clone = commons.empty_clone(self)
        clone.__dict__ = self.__dict__.copy()

        clone.metadata = copy.deepcopy(self.metadata)
        clone.children = {
            key: child.clone() for key, child in self.children.items()
        }
        for child_key, child in clone.children.items():
            setattr(clone, child_key, child)

        return clone

    def _get_at_instant(self, instant: Instant) -> ParameterNodeAtInstant:
        return ParameterNodeAtInstant(self.name, self, instant)

    def attach_to_parent(self, parent: "ParameterNode"):
        self.parent = parent



from .formulas import apply_thresholds, concat, switch
from .misc import empty_clone, stringify_array
from .rates import average_rate, marginal_rate

__all__ = ["apply_thresholds", "concat", "switch"]
__all__ = ["empty_clone", "stringify_array", *__all__]
__all__ = ["average_rate", "marginal_rate", *__all__]



from __future__ import annotations

import typing
from typing import Dict, Optional, Union

import numpy

from policyengine_core import tracers
from policyengine_core.enums import EnumArray

if typing.TYPE_CHECKING:
    from numpy.typing import ArrayLike

    Array = Union[EnumArray, ArrayLike]
    Trace = Dict[str, dict]


class FlatTrace:

    _full_tracer: tracers.FullTracer

    def __init__(self, full_tracer: tracers.FullTracer) -> None:
        self._full_tracer = full_tracer

    def key(self, node: tracers.TraceNode) -> str:
        name = node.name
        period = node.period
        return f"{name}<{period}, ({node.branch_name})>"

    def get_trace(self) -> dict:
        trace = {}

        for node in self._full_tracer.browse_trace():
            # We don't want cache read to overwrite data about the initial
            # calculation.
            #
            # We therefore use a non-overwriting update.
            trace.update(
                {
                    key: node_trace
                    for key, node_trace in self._get_flat_trace(node).items()
                    if key not in trace
                }
            )

        return trace

    def get_serialized_trace(self) -> dict:
        return {
            key: {**flat_trace, "value": self.serialize(flat_trace["value"])}
            for key, flat_trace in self.get_trace().items()
        }

    def serialize(
        self,
        value: Optional[Array],
    ) -> Union[Optional[Array], list]:
        if isinstance(value, EnumArray):
            value = value.decode_to_str()

        if isinstance(value, numpy.ndarray) and numpy.issubdtype(
            value.dtype, numpy.dtype(bytes)
        ):
            value = value.astype(numpy.dtype(str))

        if isinstance(value, numpy.ndarray):
            value = value.tolist()

        return value

    def _get_flat_trace(
        self,
        node: tracers.TraceNode,
    ) -> Trace:
        key = self.key(node)

        node_trace = {
            key: {
                "dependencies": [self.key(child) for child in node.children],
                "parameters": {
                    self.key(parameter): self.serialize(parameter.value)
                    for parameter in node.parameters
                },
                "value": node.value,
                "calculation_time": node.calculation_time(),
                "formula_time": node.formula_time(),
            },
        }

        return node_trace



from typing import Dict

import requests
from tqdm import tqdm

from policyengine_core.data.dataset import Dataset


class PublicDataset(Dataset):
    """Datasets inheriting from this class are stored on publicly accessible URLs."""

    url_by_year: Dict[int, str] = None
    """A dictionary mapping years to URLs."""

    def download(self, year: int):
        """Downloads the dataset from the given URL for the given year.

        Args:
            year (int): The year of the dataset to download.
        """
        url = self.url_by_year[year]
        # Download with progress bar
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            total_size = int(r.headers.get("content-length", 0))
            block_size = 1024
            t = tqdm(total=total_size, unit="iB", unit_scale=True)
            with open(self.file(year), "wb") as f:
                for data in r.iter_content(block_size):
                    t.update(len(data))
                    f.write(data)
            t.close()
            if total_size != 0 and t.n != total_size:
                raise Exception(
                    "Downloaded file size does not match content-length header."
                )

    def upload(self, year: int):
        raise AttributeError(
            "Uploading automatically is not supported for public datasets - upload to GitHub releases manually instead."
        )



from typing import Dict, Type

from microdf import MicroDataFrame, MicroSeries

from policyengine_core.data.dataset import Dataset
from policyengine_core.periods import Period
from policyengine_core.periods import period as get_period
from policyengine_core.periods.config import MONTH, YEAR
from policyengine_core.simulations.simulation import Simulation
from policyengine_core.types import ArrayLike


class Microsimulation(Simulation):
    """A `Simulation` whose entities use weights to represent larger populations."""

    def get_weights(
        self, variable_name: str, period: Period, map_to: str = None
    ) -> ArrayLike:
        time_period = get_period(period)
        variable = self.tax_benefit_system.get_variable(variable_name)
        entity_key = map_to or variable.entity.key
        weight_variable_name = f"{entity_key}_weight"
        weight_variable = self.tax_benefit_system.get_variable(
            weight_variable_name
        )
        weights = None

        if time_period.unit == weight_variable.definition_period:
            weights = self.calculate(
                weight_variable_name, time_period, use_weights=False
            )
        elif (time_period.unit == MONTH) and (
            weight_variable.definition_period == YEAR
        ):
            # Common use-case. To-do: implement others if needed.
            weights = self.calculate(
                weight_variable_name, time_period.this_year, use_weights=False
            )

        return weights

    def calculate(
        self,
        variable_name: str,
        period: Period = None,
        map_to: str = None,
        use_weights: bool = True,
    ) -> MicroSeries:
        if period is not None and not isinstance(period, Period):
            period = get_period(period)
        elif period is None and self.default_calculation_period is not None:
            period = get_period(self.default_calculation_period)
        values = super().calculate(variable_name, period, map_to)
        if not use_weights:
            return values
        weights = self.get_weights(variable_name, period, map_to)
        return MicroSeries(values, weights=weights)

    def calculate_add(
        self,
        variable_name: str,
        period: Period = None,
        map_to: str = None,
        use_weights: bool = True,
    ) -> MicroSeries:
        values = super().calculate_add(variable_name, period, map_to)
        if not use_weights:
            return values
        weights = self.get_weights(variable_name, period)
        return MicroSeries(values, weights=weights)

    def calculate_divide(
        self,
        variable_name: str,
        period: Period = None,
        map_to: str = None,
        use_weights: bool = True,
    ) -> MicroSeries:
        values = super().calculate_divide(variable_name, period, map_to)
        if not use_weights:
            return values
        weights = self.get_weights(variable_name, period)
        return MicroSeries(values, weights=weights)

    def calculate_dataframe(
        self,
        variable_names: list,
        period: Period = None,
        map_to: str = None,
        use_weights: bool = True,
    ) -> MicroDataFrame:
        values = super().calculate_dataframe(variable_names, period, map_to)
        if not use_weights:
            return values
        weights = self.get_weights(variable_names[0], period)
        return MicroDataFrame(values, weights=weights)



import abc
from typing import Any

from policyengine_core import periods
from policyengine_core.periods import Instant


class AtInstantLike(abc.ABC):
    """
    Base class for various types of parameters implementing the at instant protocol.
    """

    def __call__(self, instant: Instant) -> Any:
        return self.get_at_instant(instant)

    def get_at_instant(self, instant: Instant) -> Any:
        instant = str(periods.instant(instant))
        return self._get_at_instant(instant)

    @abc.abstractmethod
    def _get_at_instant(self, instant):
        ...



"""Data types and protocols used by OpenFisca Core.

The type definitions included in this sub-package are intented for
contributors, to help them better understand and document contracts
and expected behaviours.

Official Public API:
    * ``ArrayLike``
    * :attr:`.ArrayType`

Note:
    How imports are being used today::

        from policyengine_core.types import *  # Bad
        from policyengine_core.types.data_types.arrays import ArrayLike  # Bad


    The previous examples provoke cyclic dependency problems, that prevents us
    from modularizing the different components of the library, so as to make
    them easier to test and to maintain.

    How could them be used after the next major release::

        from policyengine_core.types import ArrayLike

        ArrayLike # Good: import types as publicly exposed

    .. seealso:: `PEP8#Imports`_ and `OpenFisca's Styleguide`_.

    .. _PEP8#Imports:
        https://www.python.org/dev/peps/pep-0008/#imports

    .. _OpenFisca's Styleguide:
        https://github.com/openfisca/openfisca-core/blob/master/STYLEGUIDE.md

"""

# Official Public API

from .data_types import ArrayLike, ArrayType

__all__ = ["ArrayLike", "ArrayType"]



from __future__ import annotations

import dataclasses
import typing

if typing.TYPE_CHECKING:
    import numpy

    from policyengine_core.enums import EnumArray
    from policyengine_core.periods import Period

    Array = typing.Union[EnumArray, numpy.typing.ArrayLike]
    Time = typing.Union[float, int]


@dataclasses.dataclass
class TraceNode:
    name: str
    period: Period
    branch_name: str = "default"
    parent: typing.Optional[TraceNode] = None
    children: typing.List[TraceNode] = dataclasses.field(default_factory=list)
    parameters: typing.List[TraceNode] = dataclasses.field(
        default_factory=list
    )
    value: typing.Optional[Array] = None
    start: float = 0
    end: float = 0

    def calculation_time(self, round_: bool = True) -> Time:
        result = self.end - self.start

        if round_:
            return self.round(result)

        return result

    def formula_time(self) -> float:
        children_calculation_time = sum(
            child.calculation_time(round_=False) for child in self.children
        )

        result = (
            +self.calculation_time(round_=False) - children_calculation_time
        )

        return self.round(result)

    def append_child(self, node: TraceNode) -> None:
        self.children.append(node)

    @staticmethod
    def round(time: Time) -> float:
        return float(f"{time:.4g}")  # Keep only 4 significant figures



from __future__ import annotations

from typing import Optional

import sortedcontainers

from policyengine_core.periods import Period

from .. import variables


def get_annualized_variable(
    variable: variables.Variable, annualization_period: Optional[Period] = None
) -> variables.Variable:
    """
    Returns a clone of ``variable`` that is annualized for the period ``annualization_period``.
    When annualized, a variable's formula is only called for a January calculation, and the results for other months are assumed to be identical.
    """

    def make_annual_formula(original_formula, annualization_period=None):
        def annual_formula(population, period, parameters):
            if period.start.month != 1 and (
                annualization_period is None
                or annualization_period.contains(period)
            ):
                return population(variable.name, period.this_year.first_month)
            if original_formula.__code__.co_argcount == 2:
                return original_formula(population, period)
            return original_formula(population, period, parameters)

        return annual_formula

    new_variable = variable.clone()
    new_variable.formulas = sortedcontainers.sorteddict.SortedDict(
        {
            key: make_annual_formula(formula, annualization_period)
            for key, formula in variable.formulas.items()
        }
    )

    return new_variable


def get_neutralized_variable(variable):
    """
    Return a new neutralized variable (to be used by reforms).
    A neutralized variable always returns its default value, and does not cache anything.
    """
    result = variable.clone()
    result.is_neutralized = True
    result.label = (
        "[Neutralized]"
        if variable.label is None
        else "[Neutralized] {}".format(variable.label),
    )

    return result


def _partition(dict, predicate):
    true_dict = {}
    false_dict = {}

    for key, value in dict.items():
        if predicate(key, value):
            true_dict[key] = value
        else:
            false_dict[key] = value

    return true_dict, false_dict



from typing import List, Union

from .reference import Reference


class VariableBreakdown(str):
    """
    The name of a variable with Enum type, whose possible values form the child nodes in a parameter node.
    """


class DynamicBreakdown(str):
    """
    A Python expression that evaluates to a list of child nodes in a parameter node.
    """


class ParameterNodeMetadata:
    """A collection of related parameters."""

    name: str
    """The name identifying this parameter node and its children (unless overridden). Should be snake-case and Python-safe."""

    label: str
    """The display text to use for the parameter node and its children (unless overridden). Should be short and less than a full sentence."""

    reference: List[Reference]
    """A list of references that apply to all descendants of this node."""

    breakdown: List[Union[VariableBreakdown, DynamicBreakdown]]
    """The sets defining the children, grandchildren and further descendants of this node.

    For example:

    .. code-block:: yaml

        metadata:
            breakdown:
                - region # If `region` is an Enum-type variable with possible values `[ENGLAND, WALES]` then these children will be added.
                - range(1, 7) # This code is `eval`uated to produce the list `[1, 2, 3, 4, 5, 6]` which are then added as grandchildren.
                - [True, False] # This list is added as great-grandchildren (using the same `eval` method as above).
    
    """

    propagate_metadata_to_children: bool
    """Whether to propagate metadata to children of this node. This excludes `breakdown`."""



"""
This file defines an additional variable for the modelled legislation.

A variable is a property of an Entity such as a Person, a Household…

See https://openfisca.org/doc/key-concepts/variables.html
"""

from policyengine_core.country_template.entities import Household

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
# Import the entities specifically defined for this tax and benefit system
from policyengine_core.periods import MONTH
from policyengine_core.variables import Variable


class local_town_child_allowance(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Local benefit: a fixed amount by child each month"

    def formula(famille, period, parameters):
        """
        Local benefit.

        Extensions can only add variables and parameters to the tax and benefit
        system: they cannot modify or neutralize existing ones.
        """
        nb_children = famille.nb_persons(role=Household.CHILD)
        amount_by_child = parameters(period).local_town.child_allowance.amount

        return nb_children * amount_by_child



import traceback
from typing import TYPE_CHECKING, Any, List

import numpy
from numpy.typing import ArrayLike

from policyengine_core import projectors
from policyengine_core.entities import Entity, Role
from policyengine_core.holders import Holder
from policyengine_core.periods.period_ import Period
from policyengine_core.populations import config
from policyengine_core.projectors import Projector

if TYPE_CHECKING:
    from policyengine_core.simulations import Simulation


class Population:
    def __init__(self, entity: Entity):
        self.simulation: "Simulation" = None
        self.entity = entity
        self._holders = {}
        self.count = 0
        self.ids = []

    def clone(self, simulation: "Simulation") -> "Population":
        result = Population(self.entity)
        result.simulation = simulation
        result._holders = {
            variable: holder.clone(result)
            for (variable, holder) in self._holders.items()
        }
        result.count = self.count
        result.ids = self.ids
        return result

    def empty_array(self) -> numpy.ndarray:
        return numpy.zeros(self.count)

    def filled_array(self, value: Any, dtype: Any = None) -> numpy.ndarray:
        return numpy.full(self.count, value, dtype)

    def __getattr__(self, attribute: str) -> Any:
        projector = projectors.get_projector_from_shortcut(self, attribute)
        if not projector:
            raise AttributeError(
                "You tried to use the '{}' of '{}' but that is not a known attribute.".format(
                    attribute, self.entity.key
                )
            )
        return projector

    def get_index(self, id: str) -> int:
        return self.ids.index(id)

    # Calculations

    def check_array_compatible_with_entity(self, array: numpy.ndarray) -> None:
        if not self.count == array.size:
            raise ValueError(
                "Input {} is not a valid value for the entity {} (size = {} != {} = count)".format(
                    array, self.entity.key, array.size, self.count
                )
            )

    def check_period_validity(
        self, variable_name: str, period: Period
    ) -> None:
        if period is None:
            stack = traceback.extract_stack()
            filename, line_number, function_name, line_of_code = stack[-3]
            raise ValueError(
                """
You requested computation of variable "{}", but you did not specify on which period in "{}:{}":
    {}
When you request the computation of a variable within a formula, you must always specify the period as the second parameter. The convention is to call this parameter "period". For example:
    computed_salary = person('salary', period).
See more information at <https://openfisca.org/doc/coding-the-legislation/35_periods.html#periods-in-variable-definition>.
""".format(
                    variable_name, filename, line_number, line_of_code
                )
            )

    def __call__(
        self, variable_name: str, period: Period = None, options: dict = None
    ):
        """
        Calculate the variable ``variable_name`` for the entity and the period ``period``, using the variable formula if it exists.

        Example:

        >>> person('salary', '2017-04')
        >>> array([300.])

        :returns: A numpy array containing the result of the calculation
        """
        self.entity.check_variable_defined_for_entity(variable_name)
        self.check_period_validity(variable_name, period)

        if options is None:
            options = []

        if config.ADD in options and config.DIVIDE in options:
            raise ValueError(
                "Options  config.ADD and  config.DIVIDE are incompatible (trying to compute variable {})".format(
                    variable_name
                ).encode(
                    "utf-8"
                )
            )

        from policyengine_core.simulations.microsimulation import (
            Microsimulation,
        )

        calculate_kwargs = {}
        if isinstance(self.simulation, Microsimulation):
            # Internal simulation code shouldn't use weights in order to avoid performance slowdowns.
            calculate_kwargs["use_weights"] = False

        if config.ADD in options:
            return self.simulation.calculate_add(
                variable_name, period, **calculate_kwargs
            )
        elif config.DIVIDE in options:
            return self.simulation.calculate_divide(
                variable_name, period, **calculate_kwargs
            )
        else:
            return self.simulation.calculate(
                variable_name, period, **calculate_kwargs
            )

    # Helpers

    def get_holder(self, variable_name: str) -> Holder:
        self.entity.check_variable_defined_for_entity(variable_name)
        holder = self._holders.get(variable_name)
        if holder:
            return holder
        variable = self.entity.get_variable(variable_name)
        self._holders[variable_name] = holder = Holder(variable, self)
        return holder

    def get_memory_usage(self, variables: List[str] = None):
        holders_memory_usage = {
            variable_name: holder.get_memory_usage()
            for variable_name, holder in self._holders.items()
            if variables is None or variable_name in variables
        }

        total_memory_usage = sum(
            holder_memory_usage["total_nb_bytes"]
            for holder_memory_usage in holders_memory_usage.values()
        )

        return dict(
            total_nb_bytes=total_memory_usage, by_variable=holders_memory_usage
        )

    @projectors.projectable
    def has_role(self, role: Role) -> ArrayLike:
        """
        Check if a person has a given role within its `GroupEntity`

        Example:

        >>> person.has_role(Household.CHILD)
        >>> array([False])
        """
        self.entity.check_role_validity(role)
        group_population = self.simulation.get_population(role.entity.plural)
        if role.subroles:
            return numpy.logical_or.reduce(
                [
                    group_population.members_role == subrole
                    for subrole in role.subroles
                ]
            )
        else:
            return group_population.members_role == role

    @projectors.projectable
    def value_from_partner(
        self, array: ArrayLike, entity: Entity, role: Role
    ) -> ArrayLike:
        self.check_array_compatible_with_entity(array)
        self.entity.check_role_validity(role)

        if not role.subroles or not len(role.subroles) == 2:
            raise Exception(
                "Projection to partner is only implemented for roles having exactly two subroles."
            )

        [subrole_1, subrole_2] = role.subroles
        value_subrole_1 = entity.value_from_person(array, subrole_1)
        value_subrole_2 = entity.value_from_person(array, subrole_2)

        return numpy.select(
            [self.has_role(subrole_1), self.has_role(subrole_2)],
            [value_subrole_2, value_subrole_1],
        )

    @projectors.projectable
    def get_rank(
        self, entity: Entity, criteria: ArrayLike, condition: ArrayLike = True
    ) -> ArrayLike:
        """
        Get the rank of a person within an entity according to a criteria.
        The person with rank 0 has the minimum value of criteria.
        If condition is specified, then the persons who don't respect it are not taken into account and their rank is -1.

        Example:

        >>> age = person('age', period)  # e.g [32, 34, 2, 8, 1]
        >>> person.get_rank(household, age)
        >>> [3, 4, 0, 2, 1]

        >>> is_child = person.has_role(Household.CHILD)  # [False, False, True, True, True]
        >>> person.get_rank(household, - age, condition = is_child)  # Sort in reverse order so that the eldest child gets the rank 0.
        >>> [-1, -1, 1, 0, 2]
        """

        # If entity is for instance 'person.household', we get the reference entity 'household' behind the projector
        entity = (
            entity
            if not isinstance(entity, Projector)
            else entity.reference_entity
        )

        positions = entity.members_position
        biggest_entity_size = numpy.max(positions) + 1
        filtered_criteria = numpy.where(condition, criteria, numpy.inf)
        ids = entity.members_entity_id

        # Matrix: the value in line i and column j is the value of criteria for the jth person of the ith entity
        matrix = numpy.asarray(
            [
                entity.value_nth_person(
                    k, filtered_criteria, default=numpy.inf
                )
                for k in range(biggest_entity_size)
            ]
        ).transpose()

        # We double-argsort all lines of the matrix.
        # Double-argsorting gets the rank of each value once sorted
        # For instance, if x = [3,1,6,4,0], y =  numpy.argsort(x) is [4, 1, 0, 3, 2] (because the value with index 4 is the smallest one, the value with index 1 the second smallest, etc.) and z =  numpy.argsort(y) is [2, 1, 4, 3, 0], the rank of each value.
        sorted_matrix = numpy.argsort(numpy.argsort(matrix))

        # Build the result vector by taking for each person the value in the right line (corresponding to its household id) and the right column (corresponding to its position)
        result = sorted_matrix[ids, positions]

        # Return -1 for the persons who don't respect the condition
        return numpy.where(condition, result, -1)



"""
IndividualSim and any other interfaces to intialising and running simulations on hypothetical situations. Deprecated.
"""
from functools import partial
from typing import Dict, List

import numpy as np

from policyengine_core.entities.entity import Entity
from policyengine_core.periods import period
from policyengine_core.reforms.reform import Reform
from policyengine_core.simulations import SimulationBuilder
from policyengine_core.taxbenefitsystems.tax_benefit_system import (
    TaxBenefitSystem,
)


class IndividualSim:
    """The IndividualSim class creates a simulation of tax-benefit policy
    on individually-specified entities.
    """

    tax_benefit_system: TaxBenefitSystem

    def __init__(
        self,
        reform: Reform = (),
        year: int = 2022,
        tax_benefit_system: TaxBenefitSystem = None,
    ) -> None:
        """Initialises a hypothetical simulation.

        Args:
            reform (ReformType, optional): The reform to apply. Defaults to ().
            year (int, optional): The default year input. Defaults to 2022.
        """
        self.year = year
        self.reform = reform
        self.system = tax_benefit_system or self.tax_benefit_system()
        self.sim_builder = SimulationBuilder()
        self.parametric_vary = False
        self.entities = {var.key: var for var in self.system.entities}
        self.situation_data = {
            entity.plural: {} for entity in self.system.entities
        }
        self.varying = False
        self.num_points = None
        self.group_entity_names = [
            entity.key
            for entity in self.system.entities
            if not entity.is_person
        ]

        # Add add_entity functions

        for entity in self.entities:
            setattr(
                self, f"add_{entity}", partial(self.add_data, entity=entity)
            )

    def build(self):
        if self.required_entities is not None:
            # Check for missing entities
            entities = {entity.key: entity for entity in self.system.entities}
            person_entity = list(
                filter(lambda x: x.is_person, entities.values())
            )[0]
            for entity in self.required_entities:
                entity_metadata = entities[entity]
                roles = {role.key: role for role in entities[entity].roles}
                default_role = roles[self.default_roles[entity]]
                no_entity_plural = (
                    entity_metadata.plural not in self.situation_data
                )
                if (
                    no_entity_plural
                    or len(self.situation_data[entity_metadata.plural]) == 0
                ):
                    self.situation_data[entity_metadata.plural] = {entity: {}}
                    members = []
                    for person in self.situation_data[person_entity.plural]:
                        members += [person]
                    self.situation_data[entity_metadata.plural][entity][
                        default_role.plural
                    ] = members
            # Add missing entities with specified default roles
        self.simulation = self.sim_builder.build_from_entities(
            self.system, self.situation_data
        )
        self.simulation.trace = True
        self.sim = self.simulation

    def add_data(
        self,
        entity: str = None,
        name: str = None,
        input_period: str = None,
        auto_period: str = True,
        **kwargs: dict,
    ) -> None:
        """Add an entity to the situation data.

        Args:
            entity (str, optional): The entity name. Defaults to "person".
            name (str, optional): The name of the entity instance. Defaults to None.
            input_period (str, optional): The input period for the values. Defaults to None.
            auto_period (str, optional): Whether to automatically repeat inputs onto subperiods. Defaults to True.
            kwargs (dict): A dictionary of (variable, value).
        """
        input_period = input_period or self.year
        entity_plural = self.entities[entity].plural
        if name is None:
            name = (
                entity + "_" + str(len(self.situation_data[entity_plural]) + 1)
            )
        if auto_period:
            data = {}
            for var, value in kwargs.items():
                try:
                    def_period = self.system.get_variable(
                        var
                    ).definition_period
                    if def_period in ["eternity", "year"]:
                        input_periods = [input_period]
                    else:
                        input_periods = period(input_period).get_subperiods(
                            def_period
                        )
                    data[var] = {
                        str(subperiod): value for subperiod in input_periods
                    }
                except:
                    data[var] = value
        self.situation_data[entity_plural][name] = data

    def get_entity(self, name: str) -> Entity:
        """Gets the entity type of the entity with a given name.

        Args:
            name (str): The name of the entity.

        Returns:
            Entity: The type of the entity.
        """
        entity_type = [
            entity
            for entity in self.entities.values()
            if name in self.situation_data[entity.plural]
        ][0]
        return entity_type

    def map_to(
        self, arr: np.array, entity: str, target_entity: str, how: str = None
    ):
        """Maps values from one entity to another.

        Args:
            arr (np.array): The values in their original position.
            entity (str): The source entity.
            target_entity (str): The target entity.
            how (str, optional): A function to use when mapping. Defaults to None.

        Raises:
            ValueError: If an invalid (dis)aggregation function is passed.

        Returns:
            np.array: The mapped values.
        """
        entity_pop = self.simulation.populations[entity]
        target_pop = self.simulation.populations[target_entity]
        if entity == "person" and target_entity in self.group_entity_names:
            if how and how not in (
                "sum",
                "any",
                "min",
                "max",
                "all",
                "value_from_first_person",
            ):
                raise ValueError("Not a valid function.")
            return target_pop.__getattribute__(how or "sum")(arr)
        elif entity in self.group_entity_names and target_entity == "person":
            if not how:
                return entity_pop.project(arr)
            if how == "mean":
                return entity_pop.project(arr / entity_pop.nb_persons())
        elif entity == target_entity:
            return arr
        else:
            return self.map_to(
                self.map_to(arr, entity, "person", how="mean"),
                "person",
                target_entity,
                how="sum",
            )

    def get_group(self, entity: str, name: str) -> str:
        """Gets the name of the containing entity for a named entity and group type.

        Args:
            entity (str): The group type, e.g. "household".
            name (str): The name of the entity, e.g. "person1".

        Returns:
            str: The containing entity, e.g. "household1".
        """
        containing_entity = [
            group
            for group in self.situation_data[entity.plural]
            if name in self.situation_data[entity.plural][group]["adults"]
            or name in self.situation_data[entity.plural][group]["children"]
        ][0]
        return containing_entity

    def calc(
        self,
        var: str,
        period: int = None,
        target: str = None,
        index: int = None,
        map_to: str = None,
        reform: Reform = None,
    ) -> np.array:
        """Calculates the value of a variable, executing any required formulas.

        Args:
            var (str): The variable to calculate.
            period (int, optional): The time period to calculate for. Defaults to None.
            target (str, optional): The target entity if not all entities are required. Defaults to None.
            index (int, optional): The numerical index of the target entity. Defaults to None.
            reform (reform, optional): The reform to apply. Defaults to None.

        Returns:
            np.array: The resulting values.
        """
        if not hasattr(self, "simulation"):
            self.build()

        if self.parametric_vary and reform is None:
            results = [
                self.calc(
                    var,
                    period=period,
                    target=target,
                    index=index,
                    map_to=map_to,
                    reform=reform,
                )
                for reform in self.parametric_reforms
            ]
            return np.array(results)

        if reform is not None:
            reform.apply(self)

        period = period or self.year
        entity = self.system.variables[var].entity
        try:
            result = self.simulation.calculate(var, period)
        except:
            try:
                result = self.sim.calculate_add(var, period)
            except:
                result = self.simulation.calculate_divide(var, period)
        if (
            target is not None
            and target not in self.situation_data[entity.plural]
        ):
            map_to = self.get_entity(target).key
        if map_to is not None:
            result = self.map_to(result, entity.key, map_to)
            entity = self.entities[map_to]
        if self.varying:
            result = result.reshape(
                (self.num_points, len(self.situation_data[entity.plural]))
            ).transpose()
        members = list(self.situation_data[entity.plural])
        if index is not None:
            index = min(len(members) - 1, index)
        if target is not None:
            index = members.index(target)
        if target is not None or index is not None:
            return result[index]
        return result

    def deriv(
        self,
        var: str,
        wrt: str = "employment_income",
        period: int = None,
        var_target: str = None,
        wrt_target: str = None,
    ):
        """Calculates the derivative of one variable with respect to another.

        Args:
            var (str): The target variable.
            wrt (str, optional): The varying variable. Defaults to "employment_income".
            period (int, optional): The time period to calculate over. Defaults to None.
            var_target (str, optional): The target name. Defaults to None.
            wrt_target (str, optional): The source name. Defaults to None.

        Returns:
            np.array: The derivatives as the source variable varies.
        """
        period = period or self.year
        y = self.calc(var, period=period, target=var_target)
        x = self.calc(wrt, period=period, target=wrt_target)
        try:
            y = y.squeeze()
        except:
            pass
        try:
            x = x.squeeze()
        except:
            pass
        x = x.astype(np.float32)
        y = y.astype(np.float32)
        assert (
            len(y) > 1 and len(x) > 1
        ), "Simulation must vary on an axis to calculate derivatives."
        deriv = (y[1:] - y[:-1]) / (x[1:] - x[:-1])
        deriv = np.append(deriv, deriv[-1])
        return deriv

    def reset_vary(self) -> None:
        """Removes an axis from the simulation."""
        del self.situation_data["axes"]
        self.varying = False
        self.num_points = None

    def vary(
        self,
        var: str = None,
        parameter: str = None,
        min: float = 0,
        max: float = 200000,
        step: float = 100,
        index: int = 0,
        period: int = None,
    ) -> None:
        """Adds an axis to the situation, varying one variable.

        Args:
            var (str, optional): The variable to change.
            parameter (str, optional): The parameter to vary. Defaults to None.
            min (float, optional): The minimum value. Defaults to 0.
            max (float, optional): The maximum value. Defaults to 200000.
            step (float, optional): The step size. Defaults to 100.
            index (int, optional): The specific entity index to target. Defaults to 0.
            period (int, optional): The time period. Defaults to None.
        """
        count = int((max - min) / step) + 1
        if var is not None:
            period = period or self.year
            if "axes" not in self.situation_data:
                self.situation_data["axes"] = [[]]
            self.situation_data["axes"][0] += [
                {
                    "count": count,
                    "name": var,
                    "min": min,
                    "max": max,
                    "period": period,
                    "index": index,
                }
            ]
            self.build()
            self.varying = True

            self.num_points = count
        if parameter is not None:
            # Parametric vary
            self.parametric_vary = True
            parameter_values = np.linspace(min, max, count)
            self.parametric_reforms = [
                set_parameter(parameter, value, period or "year:2022:10")
                for value in parameter_values
            ]



from __future__ import annotations

import typing

import numpy

from policyengine_core.taxscales import AmountTaxScaleLike

if typing.TYPE_CHECKING:
    NumericalArray = typing.Union[numpy.int_, numpy.float_]


class SingleAmountTaxScale(AmountTaxScaleLike):
    def calc(
        self,
        tax_base: NumericalArray,
        right: bool = False,
    ) -> numpy.float_:
        """
        Matches the input amount to a set of brackets and returns the single
        cell value that fits within that bracket.
        """
        guarded_thresholds = numpy.array(
            [-numpy.inf] + self.thresholds + [numpy.inf]
        )

        bracket_indices = numpy.digitize(
            tax_base,
            guarded_thresholds,
            right=right,
        )

        guarded_amounts = numpy.array([0] + self.amounts + [0])

        return guarded_amounts[bracket_indices - 1]



from typing import TYPE_CHECKING, Any

import numpy
from numpy.typing import ArrayLike

from policyengine_core import parameters
from policyengine_core.enums import Enum, EnumArray
from policyengine_core.errors import ParameterNotFoundError
from policyengine_core.parameters import helpers

if TYPE_CHECKING:
    from policyengine_core.parameters.parameter_node import ParameterNode


class VectorialParameterNodeAtInstant:
    """
    Parameter node of the legislation at a given instant which has been vectorized.
    Vectorized parameters allow requests such as parameters.housing_benefit[zipcode], where zipcode is a vector
    """

    @staticmethod
    def build_from_node(
        node: "ParameterNode",
    ) -> "VectorialParameterNodeAtInstant":
        VectorialParameterNodeAtInstant.check_node_vectorisable(node)
        subnodes_name = node._children.keys()
        # Recursively vectorize the children of the node
        vectorial_subnodes = tuple(
            [
                VectorialParameterNodeAtInstant.build_from_node(
                    node[subnode_name]
                ).vector
                if isinstance(
                    node[subnode_name], parameters.ParameterNodeAtInstant
                )
                else node[subnode_name]
                for subnode_name in subnodes_name
            ]
        )
        # A vectorial node is a wrapper around a numpy recarray
        # We first build the recarray
        recarray = numpy.array(
            [vectorial_subnodes],
            dtype=[
                (
                    subnode_name,
                    subnode.dtype
                    if isinstance(subnode, numpy.recarray)
                    else "float",
                )
                for (subnode_name, subnode) in zip(
                    subnodes_name, vectorial_subnodes
                )
            ],
        )

        return VectorialParameterNodeAtInstant(
            node._name, recarray.view(numpy.recarray), node._instant_str
        )

    @staticmethod
    def check_node_vectorisable(node: "ParameterNode") -> None:
        """
        Check that a node can be casted to a vectorial node, in order to be able to use fancy indexing.
        """
        MESSAGE_PART_1 = "Cannot use fancy indexing on parameter node '{}', as"
        MESSAGE_PART_3 = "To use fancy indexing on parameter node, its children must be homogenous."
        MESSAGE_PART_4 = "See more at <https://openfisca.org/doc/coding-the-legislation/legislation_parameters#computing-a-parameter-that-depends-on-a-variable-fancy-indexing>."

        def raise_key_inhomogeneity_error(
            node_with_key, node_without_key, missing_key
        ):
            message = " ".join(
                [
                    MESSAGE_PART_1,
                    "'{}' exists, but '{}' doesn't.",
                    MESSAGE_PART_3,
                    MESSAGE_PART_4,
                ]
            ).format(
                node._name,
                ".".join([node_with_key, missing_key]),
                ".".join([node_without_key, missing_key]),
            )

            raise ValueError(message)

        def raise_type_inhomogeneity_error(node_name, non_node_name):
            message = " ".join(
                [
                    MESSAGE_PART_1,
                    "'{}' is a node, but '{}' is not.",
                    MESSAGE_PART_3,
                    MESSAGE_PART_4,
                ]
            ).format(
                node._name,
                node_name,
                non_node_name,
            )

            raise ValueError(message)

        def raise_not_implemented(node_name, node_type):
            message = " ".join(
                [
                    MESSAGE_PART_1,
                    "'{}' is a '{}', and fancy indexing has not been implemented yet on this kind of parameters.",
                    MESSAGE_PART_4,
                ]
            ).format(
                node._name,
                node_name,
                node_type,
            )
            raise NotImplementedError(message)

        def extract_named_children(node):
            return {
                ".".join([node._name, key]): value
                for key, value in node._children.items()
            }

        def check_nodes_homogeneous(named_nodes):
            """
            Check than several nodes (or parameters, or baremes) have the same structure.
            """
            names = list(named_nodes.keys())
            nodes = list(named_nodes.values())
            first_node = nodes[0]
            first_name = names[0]
            if isinstance(first_node, parameters.ParameterNodeAtInstant):
                children = extract_named_children(first_node)
                for node, name in list(zip(nodes, names))[1:]:
                    if not isinstance(node, parameters.ParameterNodeAtInstant):
                        raise_type_inhomogeneity_error(first_name, name)
                    first_node_keys = first_node._children.keys()
                    node_keys = node._children.keys()
                    if not first_node_keys == node_keys:
                        missing_keys = set(first_node_keys).difference(
                            node_keys
                        )
                        if (
                            missing_keys
                        ):  # If the first_node has a key that node hasn't
                            raise_key_inhomogeneity_error(
                                first_name, name, missing_keys.pop()
                            )
                        else:  # If If the node has a key that first_node doesn't have
                            missing_key = (
                                set(node_keys)
                                .difference(first_node_keys)
                                .pop()
                            )
                            raise_key_inhomogeneity_error(
                                name, first_name, missing_key
                            )
                    children.update(extract_named_children(node))
                check_nodes_homogeneous(children)
            elif isinstance(first_node, float) or isinstance(first_node, int):
                for node, name in list(zip(nodes, names))[1:]:
                    if isinstance(node, int) or isinstance(node, float):
                        pass
                    elif isinstance(node, parameters.ParameterNodeAtInstant):
                        raise_type_inhomogeneity_error(name, first_name)
                    else:
                        raise_not_implemented(name, type(node).__name__)

            else:
                raise_not_implemented(first_name, type(first_node).__name__)

        check_nodes_homogeneous(extract_named_children(node))

    def __init__(self, name: str, vector: ArrayLike, instant_str: str):

        self.vector = vector
        self._name = name
        self._instant_str = instant_str

    def __getattr__(self, attribute: str) -> Any:
        result = getattr(self.vector, attribute)
        if isinstance(result, numpy.recarray):
            return VectorialParameterNodeAtInstant(result)
        return result

    def __getitem__(self, key: str) -> Any:
        # If the key is a string, just get the subnode
        if isinstance(key, str):
            return self.__getattr__(key)
        # If the key is a vector, e.g. ['zone_1', 'zone_2', 'zone_1']
        elif isinstance(key, numpy.ndarray):
            if not numpy.issubdtype(key.dtype, numpy.str_):
                # In case the key is not a string vector, stringify it
                if key.dtype == object and issubclass(type(key[0]), Enum):
                    enum = type(key[0])
                    key = numpy.select(
                        [key == item for item in enum],
                        [item.name for item in enum],
                    )
                elif isinstance(key, EnumArray):
                    enum = key.possible_values
                    key = numpy.select(
                        [key == item.index for item in enum],
                        [item.name for item in enum],
                    )
                else:
                    key = key.astype("str")
            names = list(
                self.dtype.names
            )  # Get all the names of the subnodes, e.g. ['zone_1', 'zone_2']
            default = numpy.full_like(
                self.vector[key[0]], numpy.nan
            )  # In case of unexpected key, we will set the corresponding value to NaN.
            conditions = [key == name for name in names]
            values = [self.vector[name] for name in names]
            result = numpy.select(conditions, values, default)
            if helpers.contains_nan(result):
                unexpected_key = (
                    set(key).difference(self.vector.dtype.names).pop()
                )
                raise ParameterNotFoundError(
                    ".".join([self._name, unexpected_key]), self._instant_str
                )

            # If the result is not a leaf, wrap the result in a vectorial node.
            if numpy.issubdtype(result.dtype, numpy.record):
                return VectorialParameterNodeAtInstant(
                    self._name, result.view(numpy.recarray), self._instant_str
                )

            return result



from typing import TYPE_CHECKING

from policyengine_core.projectors import helpers

if TYPE_CHECKING:
    from policyengine_core.populations import Population

from numpy.typing import ArrayLike


class Projector:
    reference_entity: "Population" = None
    parent: "Projector" = None

    def __getattr__(self, attribute):
        projector = helpers.get_projector_from_shortcut(
            self.reference_entity, attribute, parent=self
        )
        if projector:
            return projector

        reference_attr = getattr(self.reference_entity, attribute)
        if not hasattr(reference_attr, "projectable"):
            return reference_attr

        def projector_function(*args, **kwargs):
            result = reference_attr(*args, **kwargs)
            return self.transform_and_bubble_up(result)

        return projector_function

    def __call__(self, *args, **kwargs):
        result = self.reference_entity(*args, **kwargs)
        return self.transform_and_bubble_up(result)

    def transform_and_bubble_up(self, result):
        transformed_result = self.transform(result)
        if self.parent is None:
            return transformed_result
        else:
            return self.parent.transform_and_bubble_up(transformed_result)

    def transform(self, result: ArrayLike):
        return NotImplementedError()



import os
import traceback

import numpy

from policyengine_core import parameters, periods
from policyengine_core.errors import ParameterParsingError
from policyengine_core.parameters import config


def contains_nan(vector):
    if numpy.issubdtype(vector.dtype, numpy.record):
        return any([contains_nan(vector[name]) for name in vector.dtype.names])
    else:
        return numpy.isnan(vector).any()


def load_parameter_file(file_path, name=""):
    """
    Load parameters from a YAML file (or a directory containing YAML files).

    :returns: An instance of :class:`.ParameterNode` or :class:`.ParameterScale` or :class:`.Parameter`.
    """
    if not os.path.exists(file_path):
        raise ValueError("{} does not exist".format(file_path))
    if os.path.isdir(file_path):
        return parameters.ParameterNode(name, directory_path=file_path)
    data = _load_yaml_file(file_path)
    return _parse_child(name, data, file_path)


def _compose_name(path, child_name=None, item_name=None):
    if not path:
        return child_name
    if child_name is not None:
        return "{}.{}".format(path, child_name)
    if item_name is not None:
        return "{}[{}]".format(path, item_name)


def _load_yaml_file(file_path):
    with open(file_path, "r") as f:
        try:
            return config.yaml.load(f, Loader=config.Loader)
        except (
            config.yaml.scanner.ScannerError,
            config.yaml.parser.ParserError,
        ):
            stack_trace = traceback.format_exc()
            raise ParameterParsingError(
                "Invalid YAML. Check the traceback above for more details.",
                file_path,
                stack_trace,
            )
        except Exception:
            stack_trace = traceback.format_exc()
            raise ParameterParsingError(
                "Invalid parameter file content. Check the traceback above for more details.",
                file_path,
                stack_trace,
            )


def _parse_child(child_name, child, child_path):
    if "values" in child:
        return parameters.Parameter(child_name, child, child_path)
    elif "brackets" in child:
        return parameters.ParameterScale(child_name, child, child_path)
    elif isinstance(child, dict) and all(
        [periods.INSTANT_PATTERN.match(str(key)) for key in child.keys()]
    ):
        return parameters.Parameter(child_name, child, child_path)
    else:
        return parameters.ParameterNode(
            child_name, data=child, file_path=child_path
        )


def _validate_parameter(parameter, data, data_type=None, allowed_keys=None):
    type_map = {
        dict: "object",
        list: "array",
    }

    if data_type is not None and not isinstance(data, data_type):
        raise ParameterParsingError(
            "'{}' must be of type {}.".format(
                parameter.name, type_map[data_type]
            ),
            parameter.file_path,
        )



from __future__ import annotations

import abc
import bisect
import os
import typing

import numpy

from policyengine_core import tools
from policyengine_core.errors import EmptyArgumentError
from policyengine_core.taxscales.tax_scale_like import TaxScaleLike

if typing.TYPE_CHECKING:
    NumericalArray = typing.Union[numpy.int_, numpy.float_]


class RateTaxScaleLike(TaxScaleLike, abc.ABC):
    """
    Base class for various types of rate-based tax scales: marginal rate,
    linear average rate...
    """

    rates: typing.List

    def __init__(
        self,
        name: typing.Optional[str] = None,
        option: typing.Any = None,
        unit: typing.Any = None,
    ) -> None:
        super().__init__(name, option, unit)
        self.rates = []

    def __repr__(self) -> str:
        return tools.indent(
            os.linesep.join(
                [
                    f"- threshold: {threshold}{os.linesep}  rate: {rate}"
                    for (threshold, rate) in zip(self.thresholds, self.rates)
                ]
            )
        )

    def add_bracket(
        self,
        threshold: typing.Union[int, float],
        rate: typing.Union[int, float],
    ) -> None:
        if threshold in self.thresholds:
            i = self.thresholds.index(threshold)
            self.rates[i] += rate

        else:
            i = bisect.bisect_left(self.thresholds, threshold)
            self.thresholds.insert(i, threshold)
            self.rates.insert(i, rate)

    def multiply_rates(
        self,
        factor: float,
        inplace: bool = True,
        new_name: typing.Optional[str] = None,
    ) -> RateTaxScaleLike:
        if inplace:
            assert new_name is None

            for i, rate in enumerate(self.rates):
                self.rates[i] = rate * factor

            return self

        new_tax_scale = self.__class__(
            new_name or self.name,
            option=self.option,
            unit=self.unit,
        )

        for threshold, rate in zip(self.thresholds, self.rates):
            new_tax_scale.thresholds.append(threshold)
            new_tax_scale.rates.append(rate * factor)

        return new_tax_scale

    def multiply_thresholds(
        self,
        factor: float,
        decimals: typing.Optional[int] = None,
        inplace: bool = True,
        new_name: typing.Optional[str] = None,
    ) -> RateTaxScaleLike:
        if inplace:
            assert new_name is None

            for i, threshold in enumerate(self.thresholds):
                if decimals is not None:
                    self.thresholds[i] = numpy.around(
                        threshold * factor,
                        decimals=decimals,
                    )

                else:
                    self.thresholds[i] = threshold * factor

            return self

        new_tax_scale = self.__class__(
            new_name or self.name,
            option=self.option,
            unit=self.unit,
        )

        for threshold, rate in zip(self.thresholds, self.rates):
            if decimals is not None:
                new_tax_scale.thresholds.append(
                    numpy.around(threshold * factor, decimals=decimals),
                )
            else:
                new_tax_scale.thresholds.append(threshold * factor)

            new_tax_scale.rates.append(rate)

        return new_tax_scale

    def bracket_indices(
        self,
        tax_base: NumericalArray,
        factor: float = 1.0,
        round_decimals: typing.Optional[int] = None,
    ) -> numpy.int_:
        """
        Compute the relevant bracket indices for the given tax bases.

        :param ndarray tax_base: Array of the tax bases.
        :param float factor: Factor to apply to the thresholds.
        :param int round_decimals: Decimals to keep when rounding thresholds.

        :returns: Integer array with relevant bracket indices for the given tax
                  bases.

        For instance:

        >>> tax_scale = LinearAverageRateTaxScale()
        >>> tax_scale.add_bracket(0, 0)
        >>> tax_scale.add_bracket(100, 0.1)
        >>> tax_base = array([0, 150])
        >>> tax_scale.bracket_indices(tax_base)
        [0, 1]
        """

        if not numpy.size(numpy.array(self.thresholds)):
            raise EmptyArgumentError(
                self.__class__.__name__,
                "bracket_indices",
                "self.thresholds",
                self.thresholds,
            )

        if not numpy.size(numpy.asarray(tax_base)):
            raise EmptyArgumentError(
                self.__class__.__name__,
                "bracket_indices",
                "tax_base",
                tax_base,
            )

        base1 = numpy.tile(tax_base, (len(self.thresholds), 1)).T
        factor = numpy.ones(len(tax_base)) * factor

        # To avoid the creation of:
        #
        #   numpy.nan = 0 * numpy.inf
        #
        # We use:
        #
        #   numpy.finfo(float_).eps
        thresholds1 = numpy.outer(
            +factor + numpy.finfo(numpy.float_).eps,
            numpy.array(self.thresholds),
        )

        if round_decimals is not None:
            thresholds1 = numpy.round_(thresholds1, round_decimals)

        return (base1 - thresholds1 >= 0).sum(axis=1) - 1

    def threshold_from_tax_base(
        self,
        tax_base: NumericalArray,
    ) -> NumericalArray:
        """
        Compute the relevant thresholds for the given tax bases.

        :param: ndarray tax_base: Array of the tax bases.

        :returns: Floating array with relevant thresholds
                  for the given tax bases.

        For instance:

        >>> import numpy
        >>> from policyengine_core import taxscales
        >>> tax_scale = taxscales.MarginalRateTaxScale()
        >>> tax_scale.add_bracket(0, 0)
        >>> tax_scale.add_bracket(200, 0.1)
        >>> tax_scale.add_bracket(500, 0.25)
        >>> tax_base = numpy.array([450, 1_150, 10])
        >>> tax_scale.threshold_from_tax_base(tax_base)
        array([200, 500,   0])
        """

        return numpy.array(self.thresholds)[self.bracket_indices(tax_base)]

    def to_dict(self) -> dict:
        return {
            str(threshold): self.rates[index]
            for index, threshold in enumerate(self.thresholds)
        }
